"""Upgrade the database tables if needed."""

from datetime import datetime
from euphorie.client import model
from pkg_resources import get_distribution
from pkg_resources import parse_version
from sqlalchemy import sql
from sqlalchemy.engine.reflection import Inspector
from sys import argv
from transaction import commit
from z3c.saconfig import Session
from Zope2.App.zcml import load_config

import logging


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s -  %(message)s")
handler.setFormatter(formatter)
logger.setLevel(logging.INFO)
logger.addHandler(handler)

euphorie_version = get_distribution("euphorie").parsed_version

try:
    config = argv[1]
except IndexError:
    config = "parts/instance/etc/package-includes/999-additional-overrides.zcml"  # noqa: E501

load_config(config)
session = Session()
inspector = Inspector.from_engine(session.bind)


def execute(statement):
    """Execute the given SQL statement and commit immediately after it is
    executed."""
    logger.info(statement)
    session.execute(statement)
    session.execute("COMMIT;")


def create_missing_tables():
    """This will create the missing tables."""
    model.metadata.create_all(Session.bind, checkfirst=True)


def add_group_id_to_account():
    """A new 'group_id' column has been added to the 'account' table."""
    for column in inspector.get_columns("account"):
        if "group_id" == column["name"]:
            return
    statement = """
        ALTER TABLE account
            ADD COLUMN group_id character varying(32);
        ALTER TABLE account
            ADD CONSTRAINT account_group_id_fkey
            FOREIGN KEY (group_id)
            REFERENCES "group" (group_id) MATCH SIMPLE
            ON UPDATE NO ACTION
            ON DELETE NO ACTION;
        """
    execute(statement)


def add_brand_to_session():
    """A new 'brand' column has been added to the 'session' table."""
    for column in inspector.get_columns("session"):
        if "brand" == column["name"]:
            return

    statement = """
        ALTER TABLE session
            ADD COLUMN brand character varying(64);
        """
    execute(statement)


def add_brand_to_group():
    """A new 'brand' column has been added to the 'group' table."""
    for column in inspector.get_columns("group"):
        if "brand" == column["name"]:
            return

    statement = """
        ALTER TABLE "group"
            ADD COLUMN brand character varying(64);
        """
    execute(statement)


def add_group_id_to_session():
    """A new 'group_id' column has been added to the 'session' table."""
    for column in inspector.get_columns("session"):
        if "group_id" == column["name"]:
            return

    statement = """
        ALTER TABLE session
            ADD COLUMN group_id character varying(32);
        ALTER TABLE session
            ADD CONSTRAINT session_group_id_fkey
            FOREIGN KEY (group_id)
            REFERENCES public."group" (group_id) MATCH SIMPLE
            ON UPDATE NO ACTION
            ON DELETE NO ACTION;
        """
    execute(statement)


def add_archived_to_session():
    """A new 'archived' column has been added to the 'session' table."""
    for column in inspector.get_columns("session"):
        if "archived" == column["name"]:
            return

    statement = """
        ALTER TABLE session
            ADD COLUMN archived timestamp with time zone;
        """
    execute(statement)


def add_published_to_session():
    """A new 'published' column has been added to the 'session' table."""
    for column in inspector.get_columns("session"):
        if "published" == column["name"]:
            return

    statement = """
        ALTER TABLE session
            ADD COLUMN published timestamp with time zone;
        """
    execute(statement)


def add_last_modifier_id_to_session():
    """A new 'last_modifier_id' column has been added to the 'session'
    table."""
    for column in inspector.get_columns("session"):
        if "last_modifier_id" == column["name"]:
            return

    statement = """
        ALTER TABLE session
            ADD COLUMN last_modifier_id integer;
        ALTER TABLE session
            ADD CONSTRAINT session_last_modifier_id_fkey
            FOREIGN KEY (last_modifier_id)
            REFERENCES public.account (id) MATCH SIMPLE
            ON UPDATE NO ACTION
            ON DELETE NO ACTION;
        """
    execute(statement)


def add_last_publisher_id_to_session():
    """A new 'last_publisher_id' column has been added to the 'session'
    table."""
    for column in inspector.get_columns("session"):
        if "last_publisher_id" == column["name"]:
            return

    statement = """
        ALTER TABLE session
            ADD COLUMN last_publisher_id integer;
        ALTER TABLE session
            ADD CONSTRAINT session_last_publisher_id_fkey
            FOREIGN KEY (last_publisher_id)
            REFERENCES public.account (id) MATCH SIMPLE
            ON UPDATE NO ACTION
            ON DELETE NO ACTION;
        """
    execute(statement)


def add_custom_description_to_risk():
    """A new 'custom_description' column has been added to the 'risk' table."""
    for column in inspector.get_columns("risk"):
        if "custom_description" == column["name"]:
            return

    statement = """
        ALTER TABLE risk
            ADD COLUMN custom_description text;
        """
    execute(statement)


def hash_passwords():
    """We want the passwords stored in the account table to be encrypted."""
    accounts = session.query(model.Account).filter(
        sql.or_(
            sql.not_(model.Account.account_type == "guest"),
            model.Account.account_type == None,  # noqa: E711
        )
    )
    total = float(accounts.count())
    start = datetime.now()
    logger.info(
        "{} - {} accounts to convert".format(
            start.strftime("%Y/%m/%d %H:%M:%S"), int(total)
        )
    )
    cnt = 0
    for account in accounts:
        account.hash_password()
        cnt += 1
        if cnt % 500 == 0:
            logger.info(
                "{} - {} accounts converted ({:2.2f}%)".format(
                    datetime.now().strftime("%Y/%m/%d %H:%M:%S"),
                    cnt,
                    cnt / total * 100,
                )
            )
            logger.info(f"    {account.loginname}")
    logger.info(
        "{} accounts processed. Finished after {}".format(
            cnt,
            datetime.now() - start,
        )
    )
    commit()


def main():
    # It is always a good idea to run this one
    create_missing_tables()
    add_archived_to_session()
    if euphorie_version < parse_version("10.0.1"):
        add_group_id_to_account()
        add_brand_to_session()
        add_group_id_to_session()
        add_published_to_session()
        add_last_modifier_id_to_session()
        add_last_publisher_id_to_session()
        hash_passwords()
    if euphorie_version < parse_version("10.0.4"):
        add_brand_to_group()
    if euphorie_version < parse_version("11.0.5"):
        add_custom_description_to_risk()


if __name__ == "__main__":
    main()



from BTrees.OOBTree import OOBTree
from copy import deepcopy
from euphorie.client import MessageFactory as _
from euphorie.client.browser.base import BaseView
from euphorie.client.model import Account
from euphorie.client.model import Organisation
from euphorie.client.model import OrganisationMembership
from euphorie.client.utils import CreateEmailTo
from plone import api
from plone.memoize.view import memoize
from plone.memoize.view import memoize_contextless
from plone.protect.auto import safeWrite
from plone.protect.interfaces import IDisableCSRFProtection
from plone.scale.scale import scaleImage
from plone.uuid.interfaces import IUUIDGenerator
from Products.CMFPlone.utils import getAllowedSizes
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from sqlalchemy import sql
from textwrap import dedent
from time import time
from urllib.parse import quote
from zExceptions import NotFound
from zExceptions import Unauthorized
from zope.annotation.interfaces import IAnnotations
from zope.component import getUtility
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import logging


logger = logging.getLogger(__name__)


class OrganisationBaseView(BaseView):
    """Base class for all organisation views."""

    default_target_view = "@@organisation"

    @property
    def _known_roles(self):
        roles = [
            {
                "value": "member",
                "label": _("Member"),
            },
            # XXX commented for the moment because it is still under discussion
            # {
            #     "value": "manager",
            #     "label": _("Manager/Policy maker"),
            # },
            {
                "value": "admin",
                "label": _("Administrator"),
            },
        ]

        if self.webhelpers.use_consultancy_phase:
            roles.insert(
                1,
                {
                    "value": "consultant",
                    "label": _("label_role_consultant", default="Consultant"),
                },
            )
        return roles

    @property
    def is_training_enabled(self):
        return self.webhelpers.use_training_module

    @property
    def default_organisation_title(self):
        """Return the default title for a new organisation or an orgasination
        without a title set."""
        account = self.webhelpers.get_current_account()
        name = account.first_name or account.loginname
        return api.portal.translate(
            _(
                "default_organisation_title",
                default="Organisation of ${name}",
                mapping={"name": name},
            )
        )

    def get_organisation_title(self, organisation):
        """Return the title of the organisation."""
        if not organisation:
            return "TODO: FIXME"
        return organisation.title or self.default_organisation_title

    @property
    @memoize
    def organisation_title(self):
        """The title of the organisation bound to this account (if it exists)
        or the account login name."""
        return self.get_organisation_title(self.organisation)

    def get_member_role_id(self, organization, user):
        """Return the role of the user in the organization."""
        if organization.owner_id == user.id:
            return "owner"
        membership = (
            self.sqlsession.query(OrganisationMembership)
            .filter(
                OrganisationMembership.owner_id == organization.owner_id,
                OrganisationMembership.member_id == user.id,
            )
            .first()
        )
        if not membership:
            return None
        return membership.member_role

    @memoize_contextless
    def translate_role_id(self, role_id):
        """Return the translated role value."""
        for role in self._known_roles:
            if role["value"] == role_id:
                return api.portal.translate(role["label"])
        return role_id

    @property
    @memoize
    def organisations(self):
        account = self.webhelpers.get_current_account()
        if not account:
            return []

        organisations = []
        if not account.organisation:
            organisation = Organisation(
                owner_id=account.id, title=self.default_organisation_title
            )
            alsoProvides(self.request, IDisableCSRFProtection)
            self.sqlsession.add(organisation)
            account.organisation = organisation

        organisations.append(account.organisation)

        # Extend with the organisation the account is member of
        organisations.extend(
            sorted(
                self.sqlsession.query(Organisation)
                .join(
                    OrganisationMembership,
                    Organisation.owner_id == OrganisationMembership.owner_id,
                )
                .filter(OrganisationMembership.member_id == account.id),
                key=self.get_organisation_title,
            )
        )
        return organisations


class View(OrganisationBaseView):
    @memoize
    def get_memberships(self, organisation):
        return (
            self.sqlsession.query(Account, OrganisationMembership)
            .join(
                OrganisationMembership,
                OrganisationMembership.member_id == Account.id,
            )
            .filter(OrganisationMembership.owner_id == organisation.owner_id)
            .order_by(Account.loginname)
            .all()
        )


class PanelAddOrganisation(OrganisationBaseView):
    def handle_POST(self):
        """Handle the POST request."""
        title = self.request.form.get("title")
        if not title:
            return self.redirect(
                msg=_("You need to specify a title for the organisation.")
            )
        account = self.webhelpers.get_current_account()
        if account.organisation:
            return self.redirect(msg=_("You already have an organisation."))

        organisation = Organisation(owner_id=account.id, title=title)
        self.sqlsession.add(organisation)

        logo = self.request.form.get("logo")
        if logo:
            organisation.image_data = logo.read()
            organisation.image_filename = logo.filename
            # The scale will be recreated on the fly
            organisation.image_data_scaled = None
        return self.redirect(msg=_("Organisation added"))


@implementer(IPublishTraverse)
class PanelAddUser(OrganisationBaseView):
    """Panel to add a new user to an organisation."""

    storage_key = "euphorie.add_user_to_organisation_tokens"
    days_to_keep = 5

    organisation_id_key = "organisation_id"

    def publishTraverse(self, request, organisation_id):
        request.set(self.organisation_id_key, organisation_id)
        return self

    @property
    def role_options(self):
        """Return a list of options for the role field."""
        options = deepcopy(self._known_roles)
        options[0]["checked"] = "checked"
        return options

    @property
    @memoize_contextless
    def organisation(self):
        """Get the organisation requested, defaults to the current user
        organisation."""
        organisation_id = self.request.get(self.organisation_id_key)
        account = self.webhelpers.get_current_account()
        if not organisation_id:
            return account.organisation

        organisation = self.sqlsession.query(Organisation).get(
            self.request.get(self.organisation_id_key)
        )
        if not organisation:
            return

        # Check if the user can manage users to the current organisation
        account_id = account.id
        if organisation.owner_id == account_id:
            # The user is the owner, return the organisation
            return organisation

        if (
            self.sqlsession.query(OrganisationMembership)
            .filter(
                sql.and_(
                    OrganisationMembership.owner_id == organisation.owner_id,
                    OrganisationMembership.member_id == account_id,
                    OrganisationMembership.member_role == "admin",
                )
            )
            .count()
        ):
            # The user is a member, return the organisation
            return organisation

        raise Unauthorized("You are not allowed to edit this organisation")

    @property
    def subject(self):
        return api.portal.translate(
            _(
                "subject_add_user_to_organisation_email",
                default="Confirm OiRA membership to ’${organisation}’",
                mapping={"organisation": self.organisation_title},
            )
        )

    @property
    def body(self):
        """The mail body.

        This has to be quite simple because it is used in a mailto link.
        """
        body = _(
            "body_add_user_to_organisation_email",
            default="""
            I hereby invite you to join our organisation on OiRA.
            Once you've accepted your membership,
            we can collaborate on risk assessments.

            Please log in or register via the link below
            and accept your membership.
            This link will stay valid for ${days_to_keep} days.

            ${link}
            """,
            mapping={
                "days_to_keep": self.days_to_keep,
                "link": (
                    f"{self.context.absolute_url()}/"
                    f"@@confirm-organisation-invite/{self.token}"
                ),
            },
        )
        return dedent(api.portal.translate(body)).strip()

    @property
    @memoize_contextless
    def storage(self):
        annotations = IAnnotations(api.portal.get())
        if self.storage_key not in annotations:
            safeWrite(annotations)
            annotations[self.storage_key] = OOBTree()
        return annotations[self.storage_key]

    @property
    @memoize_contextless
    def token(self):
        storage = self.storage
        safeWrite(storage)

        # clean up expired tokens
        now = time()
        expires = now + self.days_to_keep * 86400

        for key, value in storage.items():
            if value.get("expires", 0) < now:
                storage.pop(key)

        uuid_generator = getUtility(IUUIDGenerator)
        token = uuid_generator()
        storage[token] = {
            "userid": self.webhelpers.get_current_account().id,
            "organisation_id": self.organisation.organisation_id,
            "expires": expires,
            "role": self.request.form.get("role", "member"),
        }
        return token

    def handle_POST(self):
        """Handle the POST request."""
        role = self.request.get("role")
        if role not in [_["value"] for _ in self._known_roles]:
            return self.redirect(msg=_("The role is invalid"), meg_type="error")

        # Note urllib.parse.urlencode does not work with every mailclient
        target = f"mailto:?subject={quote(self.subject)}&body={quote(self.body)}"
        return self.redirect(target)


@implementer(IPublishTraverse)
class ConfirmInvite(BaseView):
    """Confirm the invite to join an organisation."""

    storage_key = PanelAddUser.storage_key
    email_template = ViewPageTemplateFile("templates/notify-invite-accept.pt")

    def publishTraverse(self, request, token):
        self.request.set(self.storage_key, token)
        return self

    @property
    @memoize_contextless
    def storage(self):
        annotations = IAnnotations(api.portal.get())
        storage = annotations.get(self.storage_key, {})
        return storage

    @property
    @memoize_contextless
    def token_value(self):
        token = self.request.get(self.storage_key)
        if not token:
            raise NotFound("Token not found")
        return self.storage.get(token)

    @property
    @memoize_contextless
    def organisation(self):
        if not self.token_value:
            return
        return self.sqlsession.query(Organisation).get(
            self.token_value["organisation_id"]
        )

    @property
    @memoize_contextless
    def inviter(self):
        if not self.token_value:
            return None
        userid = self.token_value["userid"]
        try:
            user = self.sqlsession.query(Account).filter(Account.id == userid).first()
        except Exception:
            logger.warning("Unable to fetch account for username:")
            logger.warning(userid)
            return None
        return user

    @property
    def inviter_fullname(self):
        user = self.inviter
        if user is None:
            return ""
        return user.title

    @property
    def inviter_firstname(self):
        user = self.inviter
        if user is None:
            return ""
        return user.first_name or user.title

    @property
    def default_target_view(self):
        view = "@@organisation"
        organisation = self.organisation
        if not self.organisation:
            return view
        account = self.webhelpers.get_current_account()
        if account.organisation == organisation or (
            self.sqlsession.query(OrganisationMembership)
            .filter(
                sql.and_(
                    OrganisationMembership.owner_id == organisation.owner_id,
                    OrganisationMembership.member_id == account.id,
                )
            )
            .count()
        ):
            # The user is a member, return the organisation
            return f"{view}#org-{organisation.organisation_id}"
        return view

    @property
    def email_from_name(self):
        return api.portal.get_registry_record("plone.email_from_name")

    @property
    def email_from_address(self):
        return api.portal.get_registry_record("plone.email_from_address")

    def notify_inviter(self):
        invitee = self.webhelpers.get_current_account()
        invitee_firstname = invitee.first_name or invitee.title
        body = self.email_template(
            inviter=self.inviter_firstname,
            invitee=invitee_firstname,
            organisation=self.organisation.title,
        )
        subject = api.portal.translate(
            _(
                "subject_notify_invite_accept",
                default="Membership to your organisation accepted",
            )
        )
        mail = CreateEmailTo(
            self.email_from_name,
            self.email_from_address,
            invitee.email,
            subject,
            body,
        )

        api.portal.send_email(
            body=mail,
            recipient=self.inviter.email,
            sender=self.email_from_address,
            subject=subject,
        )
        logger.info("Sent email confirmation to %s", self.inviter.email)

    def lookup_token_and_redirect(self):
        """Check that:

        - we have the token
        - the token is valid
        - the token is not expired
        - the owner of the organisation is not adding himself
        - the user is not already a member of the organisation

        If that works, add the user to the organisation
        """
        value = self.token_value
        if not value:
            return self.redirect(
                msg=_(
                    "message_add_user_to_organisation_invalid",
                    default="We could not find the invitation you are looking for.",
                )
            )

        organisation = self.organisation
        organisation_title = organisation.title

        if value["expires"] < time():
            return self.redirect(
                msg=_(
                    "message_add_user_to_organisation_expired",
                    default=(
                        "The invitation from the ${name} "
                        "organisation is not valid anymore."
                    ),
                    mapping={"name": organisation_title},
                )
            )

        current_account = self.webhelpers.get_current_account()
        if current_account.id == organisation.owner_id:
            return self.redirect(
                msg=_(
                    "message_add_user_to_organisation_self",
                    default="You cannot add yourself to your organisation.",
                )
            )

        if (
            self.sqlsession.query(OrganisationMembership)
            .filter(
                sql.and_(
                    OrganisationMembership.owner_id == organisation.owner_id,
                    OrganisationMembership.member_id == current_account.id,
                )
            )
            .count()
        ):
            return self.redirect(
                msg=_(
                    "message_add_user_to_organisation_already_member",
                    default="You are already a member of the ${name} organisation.",
                    mapping={"name": organisation_title},
                )
            )

        obj = OrganisationMembership(
            owner_id=organisation.owner_id,
            member_id=current_account.id,
            member_role=value["role"],
        )
        alsoProvides(self.request, IDisableCSRFProtection)
        self.sqlsession.add(obj)
        self.notify_inviter()

        return self.redirect(
            msg=_(
                "message_add_user_to_organisation_success",
                default="You have been added to the ${name} organisation.",
                mapping={"name": organisation_title},
            )
        )

    def handle_POST(self):
        if self.request.form.get("submit", "") == "accept":
            self.lookup_token_and_redirect()
        else:
            return self.redirect(
                msg=_(
                    "message_add_user_to_organisation_declined",
                    default=(
                        "You have declined the invitation to the ${name} "
                        "organisation."
                    ),
                    mapping={"name": self.organisation.title},
                ),
                target=f"{self.context.absolute_url()}",
            )


@implementer(IPublishTraverse)
class PanelEditOrganisation(OrganisationBaseView):
    """Panel to edit a user organisation."""

    organisation_id_key = "organisation_id"

    def publishTraverse(self, request, token):
        self.request.set(self.organisation_id_key, token)
        return self

    @property
    @memoize_contextless
    def organisation(self):
        """Get the organisation requested, defaults to the current user
        organisation."""
        organisation_id = self.request.get(self.organisation_id_key)
        account = self.webhelpers.get_current_account()
        if not organisation_id:
            return account.organisation

        organisation = self.sqlsession.query(Organisation).get(
            self.request.get(self.organisation_id_key)
        )
        if not organisation:
            return

        # Check if the organisation should be workable for the current account
        account_id = account.id
        if organisation.owner_id == account_id:
            # The user is the owner, return the organisation
            return organisation

        if (
            self.sqlsession.query(OrganisationMembership)
            .filter(
                sql.and_(
                    OrganisationMembership.owner_id == organisation.owner_id,
                    OrganisationMembership.member_id == account_id,
                    OrganisationMembership.member_role == "admin",
                )
            )
            .count()
        ):
            # The user is a member, return the organisation
            return organisation

        raise Unauthorized("You are not allowed to edit this organisation")

    @property
    @memoize
    def has_logo(self):
        return self.organisation and self.organisation.image_data is not None

    def handle_POST(self):
        """Handle the POST request."""
        organisation = self.organisation
        if not organisation:
            organisation = Organisation(
                owner_id=self.webhelpers.get_current_account().id
            )
            self.sqlsession.add(organisation)
        organisation.title = self.request.form.get("title")
        logo_operation = self.request.form.get("logo_operation")
        if logo_operation in ("upload", "replace"):
            logo = self.request.form.get("logo")
            if logo:
                organisation.image_data = logo.read()
                organisation.image_filename = logo.filename
                # The scale will be recreated on the fly
                organisation.image_data_scaled = None
        elif logo_operation == "remove":
            organisation.image_data = None
            organisation.image_data_scaled = None
            organisation.image_filename = None
        self.redirect(target=f"{self.context.absolute_url()}/@@organisation")


@implementer(IPublishTraverse)
class OrganisationLogo(OrganisationBaseView):
    """View to serve the organisation logo."""

    organisation_id_key = "organisation_id"

    def publishTraverse(self, request, token):
        self.request.set(self.organisation_id_key, token)
        return self

    @property
    @memoize_contextless
    def organisation(self):
        """Get the organisation requested, defaults to the current user
        organisation."""
        organisation_id = self.request.get(self.organisation_id_key)
        account = self.webhelpers.get_current_account()
        if not organisation_id:
            return account.organisation

        organisation = self.sqlsession.query(Organisation).get(
            self.request.get(self.organisation_id_key)
        )
        if not organisation:
            return

        return organisation

    def get_or_create_image_scaled(self):
        """Get the image scaled."""
        organisation = self.organisation
        if organisation.image_data_scaled:
            return organisation.image_data_scaled
        scale = getAllowedSizes().get("preview", (400, 400))
        scaled_image_io = scaleImage(organisation.image_data, width=scale[0])[0]
        alsoProvides(self.request, IDisableCSRFProtection)
        if isinstance(scaled_image_io, bytes):
            organisation.image_data_scaled = scaled_image_io
        else:
            organisation.image_data_scaled = scaled_image_io.getvalue()
        return organisation.image_data_scaled

    def __call__(self):
        organisation = self.organisation

        if not organisation:
            raise NotFound("Organisation not found")

        if not organisation.image_data:
            raise NotFound("Organisation logo not found")

        self.request.response.setHeader(
            "Content-Type",
            (
                "image/png"
                if organisation.image_filename.endswith(".png")
                else "image/jpeg"
            ),
        )
        self.request.response.setHeader(
            "Content-Disposition",
            f"inline; filename={organisation.image_filename}",
        )
        return self.get_or_create_image_scaled()


@implementer(IPublishTraverse)
class MemberMoreMenu(OrganisationBaseView):
    def publishTraverse(self, request, membership_id):
        request.set("membership_id", membership_id)
        return self


@implementer(IPublishTraverse)
class PanelMemberEdit(OrganisationBaseView):
    """Panel to edit a member in the context of this organisation."""

    def publishTraverse(self, request, membership_id):
        request.set("membership_id", membership_id)
        return self

    @property
    @memoize
    def membership(self):
        """Return the role of the member in the organisation."""
        membership_id = self.request.get("membership_id")
        account = self.webhelpers.get_current_account()
        membership = (
            self.sqlsession.query(OrganisationMembership)
            .filter(
                OrganisationMembership.organisation_id == membership_id,
                sql.or_(
                    OrganisationMembership.owner_id == account.id,
                    sql.and_(
                        OrganisationMembership.member_id == account.id,
                        OrganisationMembership.member_role == "admin",
                    ),
                ),
            )
            .first()
        )
        if not membership:
            raise Unauthorized("You cannot edit this organisation")
        return membership

    @property
    @memoize
    def organisation(self):
        return (
            self.sqlsession.query(Organisation)
            .filter(Organisation.owner_id == self.membership.owner_id)
            .first()
        )

    @property
    @memoize
    def edited_member(self):
        """Return the member to edit."""
        return self.sqlsession.query(Account).get(self.membership.member_id)

    @property
    def role_options(self):
        """Return a list of options for the role field."""
        options = deepcopy(self._known_roles)
        current_value = self.membership.member_role
        for option in options:
            if option["value"] == current_value:
                option["checked"] = "checked"
        return options

    def validate(self):
        if not self.membership:
            raise NotFound("Member not found")

    def handle_POST(self):
        """Handle the POST request."""
        new_role = self.request.form.get("role")
        # Validate that the role is something allowed
        if new_role != self.membership.member_role:
            if new_role not in [_["value"] for _ in self._known_roles]:
                return self.redirect(msg=_("The role is invalid"), meg_type="error")
            self.membership.member_role = new_role
            return self.redirect(msg=_("Member updated"))
        return self.redirect()


@implementer(IPublishTraverse)
class PanelMemberRemove(OrganisationBaseView):
    """Panel to remove a member from this organisation."""

    def publishTraverse(self, request, membership_id):
        request.set("membership_id", membership_id)
        return self

    @property
    @memoize
    def membership(self):
        """Return the role of the member in the organisation."""
        membership_id = self.request.get("membership_id")
        account = self.webhelpers.get_current_account()
        membership = (
            self.sqlsession.query(OrganisationMembership)
            .filter(
                OrganisationMembership.organisation_id == membership_id,
                sql.or_(
                    OrganisationMembership.owner_id == account.id,
                    sql.and_(
                        OrganisationMembership.member_id == account.id,
                        OrganisationMembership.member_role == "admin",
                    ),
                ),
            )
            .first()
        )
        if not membership:
            raise Unauthorized("You cannot edit this organisation")
        return membership

    @property
    @memoize
    def organisation(self):
        return (
            self.sqlsession.query(Organisation)
            .filter(Organisation.owner_id == self.membership.owner_id)
            .first()
        )

    @property
    @memoize
    def member_to_remove(self):
        return self.sqlsession.query(Account).get(self.membership.member_id)

    def validate(self):
        if not self.membership:
            raise NotFound("Member not found")

    def handle_POST(self):
        """Handle the POST request."""
        self.sqlsession.delete(self.membership)
        return self.redirect(msg=_("Member removed"))



from euphorie.client.browser import publish
from euphorie.testing import EuphorieFunctionalTestCase
from plone.dexterity.utils import createContentInContainer


class PublishTests(EuphorieFunctionalTestCase):
    def createSectorSurvey(self):
        self.loginAsPortalOwner()
        self.sectors = self.portal.sectors
        createContentInContainer(
            self.sectors.nl, "euphorie.sector", checkConstraints=False, title="dining"
        )
        self.sector = self.sectors.nl.dining

        createContentInContainer(
            self.sector,
            "euphorie.surveygroup",
            checkConstraints=False,
            title="Restaurants",
        )
        self.surveygroup = self.sector.restaurants

        createContentInContainer(
            self.surveygroup, "euphorie.survey", checkConstraints=False, title="Survey"
        )
        self.survey = self.surveygroup.survey

        createContentInContainer(
            self.survey, "euphorie.risk", checkConstraints=False, title="Risk"
        )
        self.risk = self.survey["1"]
        return self.survey

    def testPublish(self):
        survey = self.createSectorSurvey()
        copy = publish.CopyToClient(survey)
        self.assertEqual("restaurants", "restaurants")
        self.assertEqual(survey.objectIds(), copy.objectIds())

        brains = survey.portal_catalog(portal_type="euphorie.risk")
        self.assertEqual(len(brains), 2)

        self.survey.manage_delObjects(["1"])
        copy = publish.CopyToClient(survey)
        self.assertEqual("restaurants", "restaurants")
        self.assertEqual(survey.objectIds(), copy.objectIds())

        # We had a problem whereby removed risks don't get uncatalogged,
        # check that this has now been fixed.
        brains = survey.portal_catalog(portal_type="euphorie.risk")
        self.assertEqual(len(brains), 0)



from plone.app.event.base import first_weekday
from plone.app.z3cform.widget import DateWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.widget import FieldWidget
from zope.component import ComponentLookupError
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.pagetemplate.interfaces import IPageTemplate


class EuphorieDateWidget(DateWidget):
    """Date widget for z3c.form in the Euphorie context: better display
    mode."""

    def render(self):
        """Render widget.

        :returns: Widget's HTML.
        :rtype: string
        """
        if self.mode != "input":
            return super().render()
        template = getMultiAdapter(
            (self.context, self.request, self.form, self.field, self),
            IPageTemplate,
            name=self.mode,
        )
        return template(self)


@implementer(IFieldWidget)
def EuphorieDateFieldWidget(field, request):
    """This is identicall to the equivalent in plone.app.z3cform but needs a
    customization to use the EuphorieDateWidget class."""
    widget = FieldWidget(field, EuphorieDateWidget(request))
    widget.pattern_options.setdefault("date", {})
    try:
        widget.pattern_options["date"]["firstDay"] = first_weekday()
    except ComponentLookupError:
        pass
    return widget



#
# repoze.atemplate documentation build configuration file
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# The contents of this file are pickled, so don't put values in the
# namespace that aren't pickleable (module imports are okay, they're
# removed automatically).
#
# All configuration values have a default value; values that are commented
# out serve to show the default value.

# import sys, os

# If your extensions are in another directory, add it here. If the
# directory is relative to the documentation root, use os.path.abspath to
# make it absolute, like shown here.
# sys.path.append(os.path.abspath('some/directory'))

import os
import re


def get_version():
    setup_py = os.path.join(os.path.dirname(__file__), os.pardir, "setup.py")
    setup = open(setup_py).read()
    release = re.search(r'version\s*=\s*([\'"])(.*)\1', setup).group(2)
    version = re.match(r"[0-9]+(.[0-9+])", release).group(0)
    return (version, release)


# General configuration
# ---------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.todo"]
#               'repoze.sphinx.autointerface' ]

# Add any paths that contain templates here, relative to this directory.
templates_path = [".templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General substitutions.
project = "euphorie"
copyright = "2009-2013, Simplon B.V. and SYSLAB.com GmbH"

# The default replacements for |version| and |release|, also used in various
# other places throughout the built documents.
#
(version, release) = get_version()

# There are two options for replacing |today|: either, you set today to
# some non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%B %d, %Y"

# List of documents that shouldn't be included in the build.
# unused_docs = []

# List of directories, relative to source directories, that shouldn't be
# searched for source files.
# exclude_dirs = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# Options for HTML output
# -----------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "default"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "euphoriedoc"


# Options for LaTeX output
# ------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'a4'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, document class [howto/manual]).
latex_documents = [
    ("index", "euphorie.tex", "Euphorie Documentation", "Simplon B.V.", "manual"),
]

# The name of an image file (relative to this directory) to place at the
# top of the title page.
latex_logo = ".static/logo.png"

# For "manual" documents, if this is true, then toplevel headings are
# parts, not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True


# Options for extras
# ------------------
todo_include_todos = True



from euphorie.content.browser.export import ExportSurvey
from euphorie.content.browser.upload import NSMAP
from euphorie.content.module import Module
from euphorie.content.profilequestion import ProfileQuestion
from euphorie.content.risk import Risk
from euphorie.content.solution import Solution
from euphorie.content.survey import Survey
from euphorie.content.surveygroup import SurveyGroup
from euphorie.content.training_question import TrainingQuestion
from euphorie.testing import EuphorieIntegrationTestCase
from io import BytesIO
from lxml import etree
from plone.base.utils import safe_text
from zipfile import ZipFile
from zope.publisher.browser import TestRequest


class MockImage:
    def __init__(self, data, filename=None, contentType=None):
        self.data = data
        self.filename = filename
        self.contentType = contentType


class ExportSurveyTests(EuphorieIntegrationTestCase):
    def root(self):
        return etree.Element("root", nsmap=NSMAP)

    def testImage_Minimal(self):
        image = MockImage("hot stuff here")
        root = self.root()
        view = ExportSurvey(None, None)
        view.include_images = True
        node = view.exportImage(root, image)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(node, pretty_print=True)),
            '<image xmlns="http://xml.simplon.biz/euphorie/survey/1.0">'
            "aG90IHN0dWZmIGhlcmU=\n</image>\n",
        )

    def testImage_Full(self):
        image = MockImage("hot stuff here", "test.gif", "image/gif")
        root = self.root()
        view = ExportSurvey(None, None)
        view.include_images = True
        image = view.exportImage(root, image, "Captiøn")
        self.assertEqual(
            safe_text(etree.tostring(image, pretty_print=True)),
            '<image xmlns="http://xml.simplon.biz/euphorie/survey/1.0" '
            'content-type="image/gif" filename="test.gif" '
            'caption="Capti&#xF8;n">aG90IHN0dWZmIGhlcmU=\n'
            "</image>\n",
        )

    def testSolution_Minimal(self):
        solution = Solution()
        solution.description = "<p>Test description</p>"
        solution.action = "Sample action plan"
        solution.requirements = None
        root = self.root()
        view = ExportSurvey(None, None)
        node = view.exportSolution(root, solution)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(node, pretty_print=True)),
            '<solution xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <description>&lt;p&gt;Test description&lt;/p&gt;"
            "</description>\n"
            "  <action>Sample action plan</action>\n"
            "</solution>\n",
        )

    def testSolution_Complete(self):
        solution = Solution()
        solution.description = "<p>Tést description</p>"
        solution.action = "Sample actiøn plan"
        solution.requirements = "Requîrements"
        root = self.root()
        view = ExportSurvey(None, None)
        node = view.exportSolution(root, solution)
        self.assertEqual(
            safe_text(etree.tostring(node, pretty_print=True)),
            '<solution xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <description>&lt;p&gt;T&#233;st description&lt;/p&gt;"
            "</description>\n"
            "  <action>Sample acti&#248;n plan</action>\n"
            "  <requirements>Requ&#238;rements</requirements>\n"
            "</solution>\n",
        )

    def testRisk_Minimal(self):
        risk = Risk()
        risk.type = "top5"
        risk.title = "Can your windows be locked?"
        risk.problem_description = "Not all your windows can be locked"
        risk.description = "<p>Locking windows is critical.</p>"
        risk.legal_reference = None
        risk.show_notapplicable = False
        root = self.root()
        view = ExportSurvey(None, None)
        node = view.exportRisk(root, risk)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <risk type="top5">\n'
            "    <title>Can your windows be locked?</title>\n"
            "    <problem-description>Not all your windows can be locked"
            "</problem-description>\n"
            "    <description>&lt;p&gt;Locking windows is "
            "critical.&lt;/p&gt;</description>\n"
            "    <show-not-applicable>false</show-not-applicable>\n"
            "  </risk>\n"
            "</root>\n",
        )

    def testRisk_DirectEvaluation(self):
        risk = Risk()
        risk.type = "risk"
        risk.title = "Can your windows be locked?"
        risk.problem_description = "Not all your windows can be locked"
        risk.description = "<p>Locking windows is critical.</p>"
        risk.legal_reference = None
        risk.show_notapplicable = True
        risk.evaluation_method = "direct"
        risk.default_priority = "low"
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportRisk(root, risk)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <risk type="risk">\n'
            "    <title>Can your windows be locked?</title>\n"
            "    <problem-description>Not all your windows can be "
            "locked</problem-description>\n"
            "    <description>&lt;p&gt;Locking windows is "
            "critical.&lt;/p&gt;</description>\n"
            "    <show-not-applicable>true</show-not-applicable>\n"
            '    <evaluation-method default-priority="low">direct'
            "</evaluation-method>\n"
            "  </risk>\n"
            "</root>\n",
        )

    def testRisk_CalculatedEvaluation(self):
        risk = Risk()
        risk.type = "risk"
        risk.title = "Can your windows be locked?"
        risk.problem_description = "Not all your windows can be locked"
        risk.description = "<p>Locking windows is critical.</p>"
        risk.legal_reference = None
        risk.show_notapplicable = True
        risk.evaluation_method = "calculated"
        risk.default_probability = 1
        risk.default_frequency = 4
        risk.default_effect = 0
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportRisk(root, risk)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <risk type="risk">\n'
            "    <title>Can your windows be locked?</title>\n"
            "    <problem-description>Not all your windows can be "
            "locked</problem-description>\n"
            "    <description>&lt;p&gt;Locking windows is critical."
            "&lt;/p&gt;</description>\n"
            "    <show-not-applicable>true</show-not-applicable>\n"
            '    <evaluation-method default-probability="small" '
            'default-frequency="regular">calculated</evaluation-method>\n'
            "  </risk>\n"
            "</root>\n",
        )

    def testRisk_LegalReferenceNoText(self):
        risk = Risk()
        risk.type = "top5"
        risk.title = "Can your windows be locked?"
        risk.problem_description = "Not all your windows can be locked"
        risk.description = "<p>Locking windows is critical.</p>"
        risk.legal_reference = "<p><br/></p>"
        risk.show_notapplicable = False
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportRisk(root, risk)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <risk type="top5">\n'
            "    <title>Can your windows be locked?</title>\n"
            "    <problem-description>Not all your windows can be locked"
            "</problem-description>\n"
            "    <description>&lt;p&gt;Locking windows is critical."
            "&lt;/p&gt;</description>\n"
            "    <show-not-applicable>false</show-not-applicable>\n"
            "  </risk>\n"
            "</root>\n",
        )

    def testRisk_TwoImages(self):
        risk = Risk()
        risk.type = "top5"
        risk.title = "Can your windows be locked?"
        risk.problem_description = "Not all your windows can be locked"
        risk.description = "<p>Locking windows is critical.</p>"
        risk.legal_reference = None
        risk.show_notapplicable = False
        risk.image = MockImage("hot stuff here")
        risk.caption = "Image caption 1"
        risk.image2 = MockImage("hot stuff here")
        risk.caption2 = "Image caption 2"
        root = self.root()
        view = ExportSurvey(None, None)
        view.include_images = True
        view.exportRisk(root, risk)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <risk type="top5">\n'
            "    <title>Can your windows be locked?</title>\n"
            "    <problem-description>Not all your windows can be locked"
            "</problem-description>\n"
            "    <description>&lt;p&gt;Locking windows is critical."
            "&lt;/p&gt;</description>\n"
            "    <show-not-applicable>false</show-not-applicable>\n"
            '    <image caption="Image caption 1">aG90IHN0dWZmIGhlcmU=\n'
            "</image>\n"
            '    <image caption="Image caption 2">aG90IHN0dWZmIGhlcmU=\n'
            "</image>\n"
            "  </risk>\n"
            "</root>\n",
        )

    def testRisk_WithSolution(self):
        risk = Risk()
        risk.type = "top5"
        risk.title = "Can your windows be locked?"
        risk.problem_description = "Not all your windows can be locked"
        risk.description = "<p>Locking windows is critical.</p>"
        risk.legal_reference = None
        risk.show_notapplicable = False
        solution = Solution()
        solution.description = "<p>Test description</p>"
        solution.action = "Sample action plan"
        solution.requirements = None
        risk._setOb("1", solution)
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportRisk(root, risk)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <risk type="top5">\n'
            "    <title>Can your windows be locked?</title>\n"
            "    <problem-description>Not all your windows can be locked"
            "</problem-description>\n"
            "    <description>&lt;p&gt;Locking windows is critical."
            "&lt;/p&gt;</description>\n"
            "    <show-not-applicable>false</show-not-applicable>\n"
            "    <solutions>\n"
            "      <solution>\n"
            "        <description>&lt;p&gt;Test description&lt;/p&gt;"
            "</description>\n"
            "        <action>Sample action plan</action>\n"
            "      </solution>\n"
            "    </solutions>\n"
            "  </risk>\n"
            "</root>\n",
        )

    def testModule_Minimal(self):
        module = Module()
        module.title = "Office buildings"
        module.optional = False
        module.solution_direction = None
        root = self.root()
        view = ExportSurvey(None, None)
        node = view.exportModule(root, module)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <module optional="false">\n'
            "    <title>Office buildings</title>\n"
            "  </module>\n"
            "</root>\n",
        )

    def testModule_with_description(self):
        module = Module()
        module.title = "Office buildings"
        module.description = "<p>Owning property brings risks.</p>"
        module.solution_direction = None
        module.optional = False
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportModule(root, module)
        xml = safe_text(etree.tostring(root, pretty_print=True))
        self.assertTrue(
            "<description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>\n" in xml
        )

    def testModule_Optional(self):
        module = Module()
        module.title = "Office buildings"
        module.description = "<p>Owning property brings risks.</p>"
        module.optional = True
        module.question = "Do you have an office building?"
        module.solution_direction = None
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportModule(root, module)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <module optional="true">\n'
            "    <title>Office buildings</title>\n"
            "    <description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>\n"
            "    <question>Do you have an office building?</question>\n"
            "  </module>\n"
            "</root>\n",
        )

    def testModule_SolutionDirectionNoText(self):
        module = Module()
        module.title = "Office buildings"
        module.description = "<p>Owning property brings risks.</p>"
        module.optional = False
        module.solution_direction = "<p><br/></p>"
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportModule(root, module)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <module optional="false">\n'
            "    <title>Office buildings</title>\n"
            "    <description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>\n"
            "  </module>\n"
            "</root>\n",
        )

    def testModule_Image(self):
        module = Module()
        module.title = "Office buildings"
        module.description = "<p>Owning property brings risks.</p>"
        module.optional = False
        module.solution_direction = None
        module.image = MockImage("hot stuff here")
        root = self.root()
        view = ExportSurvey(None, None)
        view.include_images = True
        view.exportModule(root, module)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <module optional="false">\n'
            "    <title>Office buildings</title>\n"
            "    <description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>\n"
            "    <image>aG90IHN0dWZmIGhlcmU=\n"
            "</image>\n"
            "  </module>\n"
            "</root>\n",
        )

    def testModule_WithRisk(self):
        module = Module()
        module.title = "Office buildings"
        module.description = "<p>Owning property brings risks.</p>"
        module.optional = False
        module.solution_direction = None
        risk = Risk()
        risk.type = "top5"
        risk.title = "Can your windows be locked?"
        risk.problem_description = "Not all your windows can be locked"
        risk.description = "<p>Locking windows is critical.</p>"
        risk.legal_reference = None
        risk.show_notapplicable = False
        module._setOb("1", risk)
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportModule(root, module)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <module optional="false">\n'
            "    <title>Office buildings</title>\n"
            "    <description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>\n"
            '    <risk type="top5">\n'
            "      <title>Can your windows be locked?</title>\n"
            "      <problem-description>Not all your windows can be "
            "locked</problem-description>\n"
            "      <description>&lt;p&gt;Locking windows is critical."
            "&lt;/p&gt;</description>\n"
            "      <show-not-applicable>false</show-not-applicable>\n"
            "    </risk>\n"
            "  </module>\n"
            "</root>\n",
        )

    def testModule_WithSubModule(self):
        module = Module()
        module.title = "Office buildings"
        module.description = "<p>Owning property brings risks.</p>"
        module.optional = False
        module.solution_direction = None
        submodule = Module()
        submodule.title = "Parking"
        submodule.description = "<p>All about parking garages.</p>"
        submodule.optional = False
        submodule.solution_direction = None
        module._setOb("1", submodule)
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportModule(root, module)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            '  <module optional="false">\n'
            "    <title>Office buildings</title>\n"
            "    <description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>\n"
            '    <module optional="false">\n'
            "      <title>Parking</title>\n"
            "      <description>&lt;p&gt;All about parking garages."
            "&lt;/p&gt;</description>\n"
            "    </module>\n"
            "  </module>\n"
            "</root>\n",
        )

    def testProfileQuestion_Minimal(self):
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.question = "Do you have an office building?"
        profile.use_location_question = False
        root = self.root()
        view = ExportSurvey(None, None)
        node = view.exportProfileQuestion(root, profile)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <profile-question>\n"
            "    <title>Office buildings</title>\n"
            "    <question>Do you have an office building?</question>\n"
            "    <use-location-question>false</use-location-question>\n"
            "  </profile-question>\n"
            "</root>\n",
        )

    def testProfileQuestion_LocationQuestions(self):
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.question = "Do you have an office building?"
        profile.label_multiple_present = "Do you have more than one building?"
        profile.label_single_occurance = "Enter the name of your building."
        profile.label_multiple_occurances = (
            "Enter the names of each of your buildings."  # noqa
        )
        root = self.root()
        view = ExportSurvey(None, None)
        node = view.exportProfileQuestion(root, profile)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <profile-question>\n"
            "    <title>Office buildings</title>\n"
            "    <question>Do you have an office building?</question>\n"
            "    <label-multiple-present>Do you have more than one building"
            "?</label-multiple-present>\n"
            "    <label-single-occurance>Enter the name of your building."
            "</label-single-occurance>\n"
            "    <label-multiple-occurances>Enter the names of each of your"
            " buildings.</label-multiple-occurances>\n"
            "    <use-location-question>true</use-location-question>\n"
            "  </profile-question>\n"
            "</root>\n",
        )

    def testProfileQuestion_with_description(self):
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.description = "<p>Owning property brings risks.</p>"
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportProfileQuestion(root, profile)
        xml = safe_text(etree.tostring(root, pretty_print=True))
        self.assertTrue(
            "<description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>" in xml
        )

    def testProfileQuestion_WithoutQuestion(self):
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.description = "<p>Owning property brings risks.</p>"
        profile.use_location_question = False
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportProfileQuestion(root, profile)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <profile-question>\n"
            "    <title>Office buildings</title>\n"
            "    <question>Office buildings</question>\n"
            "    <description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>\n"
            "    <use-location-question>false</use-location-question>\n"
            "  </profile-question>\n"
            "</root>\n",
        )

    def testProfileQuestion_WithRisk(self):
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.question = "Do you have an office buildings?"
        profile.description = "<p>Owning property brings risks.</p>"
        risk = Risk()
        risk.type = "top5"
        risk.title = "Can your windows be locked?"
        risk.problem_description = "Not all your windows can be locked"
        risk.description = "<p>Locking windows is critical.</p>"
        risk.legal_reference = None
        risk.show_notapplicable = False
        profile._setOb("1", risk)
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportProfileQuestion(root, profile)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <profile-question>\n"
            "    <title>Office buildings</title>\n"
            "    <question>Do you have an office buildings?</question>\n"
            "    <description>&lt;p&gt;Owning property brings risks."
            "&lt;/p&gt;</description>\n"
            "    <use-location-question>true</use-location-question>\n"
            '    <risk type="top5">\n'
            "      <title>Can your windows be locked?</title>\n"
            "      <problem-description>Not all your windows can be "
            "locked</problem-description>\n"
            "      <description>&lt;p&gt;Locking windows is critical."
            "&lt;/p&gt;</description>\n"
            "      <show-not-applicable>false</show-not-applicable>\n"
            "    </risk>\n"
            "  </profile-question>\n"
            "</root>\n",
        )

    def testProfileQuestion_WithModule(self):
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.question = "Do you have an office buildings?"
        profile.description = "<p>Owning property brings risks.</p>"
        module = Module()
        module.title = "Office buildings"
        module.description = "<p>Owning property brings risks.</p>"
        module.optional = False
        module.solution_direction = None
        profile._setOb("1", module)
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportProfileQuestion(root, profile)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <profile-question>\n"
            "    <title>Office buildings</title>\n"
            "    <question>Do you have an office buildings?</question>\n"
            "    <description>&lt;p&gt;Owning property brings "
            "risks.&lt;/p&gt;</description>\n"
            "    <use-location-question>true</use-location-question>\n"
            '    <module optional="false">\n'
            "      <title>Office buildings</title>\n"
            "      <description>&lt;p&gt;Owning property brings "
            "risks.&lt;/p&gt;</description>\n"
            "    </module>\n"
            "  </profile-question>\n"
            "</root>\n",
        )

    def testTrainingQuestion(self):
        training_question = TrainingQuestion()
        training_question.title = "What is the answer?"
        training_question.right_answer = "Yes"
        training_question.wrong_answer_1 = "No"
        training_question.wrong_answer_2 = "Maybe"
        root = self.root()
        view = ExportSurvey(None, None)
        node = view.exportTrainingQuestion(root, training_question)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <training_question>\n"
            "    <title>What is the answer?</title>\n"
            "    <right_answer>Yes</right_answer>\n"
            "    <wrong_answer_1>No</wrong_answer_1>\n"
            "    <wrong_answer_2>Maybe</wrong_answer_2>\n"
            "  </training_question>\n"
            "</root>\n",
        )

    def testSurvey_Minimal(self):
        surveygroup = SurveyGroup()
        surveygroup.title = "Generic sector"
        surveygroup.evaluation_algorithm = "french"
        surveygroup._setOb("standard", Survey())
        survey = surveygroup["standard"]  # Acquisition wrap
        survey.title = "Standard"
        survey.introduction = None
        survey.classification_code = None
        survey.evaluation_optional = False
        survey.language = "en-GB"
        root = self.root()
        view = ExportSurvey(None, None)
        node = view.exportSurvey(root, survey)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <survey>\n"
            "    <title>Generic sector</title>\n"
            "    <language>en-GB</language>\n"
            "    <tool_type>classic</tool_type>\n"
            "    <measures_text_handling>full</measures_text_handling>\n"
            "    <integrated_action_plan>false</integrated_action_plan>\n"
            "    <evaluation-algorithm>french</evaluation-algorithm>\n"
            "    <evaluation-optional>false</evaluation-optional>\n"
            "  </survey>\n"
            "</root>\n",
        )

    def testSurvey_IntroductionNoText(self):
        surveygroup = SurveyGroup()
        surveygroup.title = "Generic sector"
        surveygroup._setOb("standard", Survey())
        survey = surveygroup["standard"]  # Acquisition wrap
        survey.title = "Standard"
        survey.introduction = "<p><br/></p>"
        survey.classification_code = None
        survey.evaluation_optional = False
        survey.language = "en-GB"
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportSurvey(root, survey)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <survey>\n"
            "    <title>Generic sector</title>\n"
            "    <language>en-GB</language>\n"
            "    <tool_type>classic</tool_type>\n"
            "    <measures_text_handling>full</measures_text_handling>\n"
            "    <integrated_action_plan>false</integrated_action_plan>\n"
            "    <evaluation-algorithm>kinney</evaluation-algorithm>\n"
            "    <evaluation-optional>false</evaluation-optional>\n"
            "  </survey>\n"
            "</root>\n",
        )

    def testSurvey_WithProfileQuestion(self):
        surveygroup = SurveyGroup()
        surveygroup.title = "Generic sector"
        surveygroup._setOb("standard", Survey())
        survey = surveygroup["standard"]  # Acquisition wrap
        survey.title = "Generic sector"
        survey.introduction = None
        survey.classification_code = None
        survey.evaluation_optional = False
        survey.language = "en-GB"
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.question = "Do you have an office buildings?"
        profile.description = "<p>Owning property brings risks.</p>"
        profile.type = "optional"
        survey._setOb("1", profile)
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportSurvey(root, survey)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <survey>\n"
            "    <title>Generic sector</title>\n"
            "    <language>en-GB</language>\n"
            "    <tool_type>classic</tool_type>\n"
            "    <measures_text_handling>full</measures_text_handling>\n"
            "    <integrated_action_plan>false</integrated_action_plan>\n"
            "    <evaluation-algorithm>kinney</evaluation-algorithm>\n"
            "    <evaluation-optional>false</evaluation-optional>\n"
            "    <profile-question>\n"
            "      <title>Office buildings</title>\n"
            "      <question>Do you have an office buildings?</question>\n"
            "      <description>&lt;p&gt;Owning property brings "
            "risks.&lt;/p&gt;</description>\n"
            "      <use-location-question>true</use-location-question>\n"
            "    </profile-question>\n"
            "  </survey>\n"
            "</root>\n",
        )

    def testSurvey_WithModule(self):
        surveygroup = SurveyGroup()
        surveygroup.title = "Generic sector"
        surveygroup._setOb("standard", Survey())
        survey = surveygroup["standard"]  # Acquisition wrap
        survey.title = "Generic sector"
        survey.introduction = None
        survey.classification_code = None
        survey.evaluation_optional = False
        survey.language = "en-GB"
        module = Module()
        module.title = "Office buildings"
        module.description = "<p>Owning property brings risks.</p>"
        module.optional = False
        module.solution_direction = None
        survey._setOb("1", module)
        root = self.root()
        view = ExportSurvey(None, None)
        view.exportSurvey(root, survey)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <survey>\n"
            "    <title>Generic sector</title>\n"
            "    <language>en-GB</language>\n"
            "    <tool_type>classic</tool_type>\n"
            "    <measures_text_handling>full</measures_text_handling>\n"
            "    <integrated_action_plan>false</integrated_action_plan>\n"
            "    <evaluation-algorithm>kinney</evaluation-algorithm>\n"
            "    <evaluation-optional>false</evaluation-optional>\n"
            '    <module optional="false">\n'
            "      <title>Office buildings</title>\n"
            "      <description>&lt;p&gt;Owning property brings "
            "risks.&lt;/p&gt;</description>\n"
            "    </module>\n"
            "  </survey>\n"
            "</root>\n",
        )

    def testRender(self):
        surveygroup = SurveyGroup()
        surveygroup.id = "mysector"
        surveygroup.title = "Generic sector"
        surveygroup._setOb("standard", Survey())
        survey = surveygroup["standard"]  # Acquisition wrap
        survey.id = "dummy"
        survey.title = "Standard"
        survey.introduction = None
        survey.classification_code = None
        survey.evaluation_optional = False
        survey.integrated_action_plan = True
        survey.language = "en-GB"
        survey.tool_type = "existing_measures"
        view = ExportSurvey(survey, TestRequest())
        output = view.render_output().decode("utf-8")
        response = view.request.response
        self.assertEqual(response.getHeader("Content-Type"), "text/xml")
        self.assertEqual(
            response.getHeader("Content-Disposition"),
            'attachment; filename="mysector.xml"',
        )
        self.assertEqual(
            output,
            "<?xml version='1.0' encoding='utf-8'?>\n"
            '<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <survey>\n"
            "    <title>Generic sector</title>\n"
            "    <language>en-GB</language>\n"
            "    <tool_type>existing_measures</tool_type>\n"
            "    <measures_text_handling>full</measures_text_handling>\n"
            "    <integrated_action_plan>true</integrated_action_plan>\n"
            "    <evaluation-algorithm>kinney</evaluation-algorithm>\n"
            "    <evaluation-optional>false</evaluation-optional>\n"
            "  </survey>\n"
            "</sector>\n",
        )

    def testRenderZip(self):
        surveygroup = SurveyGroup()
        surveygroup.id = "mysector"
        surveygroup.title = "Generic sector"
        surveygroup._setOb("standard", Survey())
        survey = surveygroup["standard"]  # Acquisition wrap
        survey.id = "dummy"
        survey.title = "Standard"
        survey.introduction = "Wie grün sind deine Blätter?"
        survey.classification_code = None
        survey.evaluation_optional = False
        survey.integrated_action_plan = True
        survey.language = "en-GB"
        survey.tool_type = "existing_measures"
        view = ExportSurvey(survey, TestRequest())
        view.export_as_plain_text = True
        output = view.render_output()
        zip_file = ZipFile(BytesIO(output), mode="r")
        response = view.request.response
        self.assertEqual(response.getHeader("Content-Type"), "application/zip")
        self.assertEqual(
            response.getHeader("Content-Disposition"),
            'attachment; filename="mysector.zip"',
        )
        self.assertEqual(
            zip_file.namelist(), ["mysector_word_count.txt", "mysector.xml"]
        )
        with zip_file.open("mysector_word_count.txt") as txt_file:
            txt = txt_file.read().decode("utf-8").strip()
            self.assertEqual(txt, "Generic sector Wie grün sind deine Blätter?")
        with zip_file.open("mysector.xml") as xml_file:
            xml = xml_file.read().decode("utf-8")
            self.assertEqual(
                xml,
                "<?xml version='1.0' encoding='utf-8'?>\n"
                '<sector xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
                "  <survey>\n"
                "    <title>Generic sector</title>\n"
                "    <introduction>Wie grün sind deine Blätter?</introduction>\n"
                "    <language>en-GB</language>\n"
                "    <tool_type>existing_measures</tool_type>\n"
                "    <measures_text_handling>full</measures_text_handling>\n"
                "    <integrated_action_plan>true</integrated_action_plan>\n"
                "    <evaluation-algorithm>kinney</evaluation-algorithm>\n"
                "    <evaluation-optional>false</evaluation-optional>\n"
                "  </survey>\n"
                "</sector>\n",
            )

    def testEtranslateRenderWithHTML(self):
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.question = "Do you have an office building?"
        profile.description = "<p>No</p><p>container</p>"
        root = self.root()
        view = ExportSurvey(None, None)
        view.is_etranslate_compatible = True
        node = view.exportProfileQuestion(root, profile)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <profile-question>\n"
            "    <title>Office buildings</title>\n"
            "    <question>Do you have an office building?</question>\n"
            "    <description>\n"
            "      <p>No</p>\n"
            "      <p>container</p>\n"
            "    </description>\n"
            "    <use-location-question>true</use-location-question>\n"
            "  </profile-question>\n"
            "</root>\n",
        )

    def testEtranslateRenderWithoutHTML(self):
        profile = ProfileQuestion()
        profile.title = "Office buildings"
        profile.question = "Do you have an office building?"
        profile.description = "Plain text"
        root = self.root()
        view = ExportSurvey(None, None)
        view.is_etranslate_compatible = True
        node = view.exportProfileQuestion(root, profile)
        self.assertTrue(node in root)
        self.assertEqual(
            safe_text(etree.tostring(root, pretty_print=True)),
            '<root xmlns="http://xml.simplon.biz/euphorie/survey/1.0">\n'
            "  <profile-question>\n"
            "    <title>Office buildings</title>\n"
            "    <question>Do you have an office building?</question>\n"
            "    <description>Plain text</description>\n"
            "    <use-location-question>true</use-location-question>\n"
            "  </profile-question>\n"
            "</root>\n",
        )



from ftw.upgrade import UpgradeStep
from plone import api
from plone.app.upgrade.utils import update_catalog_metadata
from plone.uuid.handlers import addAttributeUUID
from plone.uuid.interfaces import ATTRIBUTE_NAME
from plone.uuid.interfaces import IUUID
from Products.ZCatalog.ProgressHandler import ZLogHandler
from zope.globalrequest import getRequest

import logging
import os
import transaction


logger = logging.getLogger(__name__)
# By default we do intermediate commits, to have less chance of ConflictErrors.
# Set the environment variable EUPHORIE_DISABLE_INTERMEDIATE_COMMITS=1
# to disable intermediate commits.
EUPHORIE_DISABLE_INTERMEDIATE_COMMITS = bool(
    int(os.environ.get("EUPHORIE_DISABLE_INTERMEDIATE_COMMITS", 0))
)


class FixUidIndex(UpgradeStep):
    """Fix UID index inconsistencies.

    The 20251006201600_ensure_uid upgrade step ensured that all content
    has a UID assigned to it.  But that was not enough to fix all problems.
    So we need to take over some more code from this script:
    https://github.com/zestsoftware/plonescripts/blob/master/fix_uid_index.py

    That script first checks paths in actual_catalog.uids.keys(), but that did
    not surface any problems in our case, so we skip that part here.

    Then it really checks the UID index.  This has two internal BTrees
    that can get out of sync:

        1) _index: UID -> doc id
        2) _unindex: doc id -> UID

    And indeed we see problems there in a test site:

        Number of _index uid keys:      198801, unique: 198801
        Number of _index doc id values: 198801, unique: 198801
        Number of _unindex doc id keys: 419339, unique: 419339
        Number of _unindex uid values:  419339, unique: 245707

    Normally, all these numbers should all be the same.
    This is not the case for us.

    The script tries to register intids for all objects that are missing them,
    but we don't have that problem.  So we skip that part.
    """

    def is_uid_index_consistent(self):
        """Check if the UID index is consistent.

        Returns True if consistent, False if not.
        """
        catalog = api.portal.get_tool("portal_catalog")
        index = catalog.Indexes["UID"]
        _index_keys = index._index.keys()
        _index_values = index._index.values()
        _unindex_keys = index._unindex.keys()
        _unindex_values = index._unindex.values()
        logger.info(
            "Number of _index uid keys:      %d, unique: %d",
            len(_index_keys),
            len(set(_index_keys)),
        )
        logger.info(
            "Number of _index doc id values: %d, unique: %d",
            len(_index_values),
            len(set(_index_values)),
        )
        logger.info(
            "Number of _unindex doc id keys: %d, unique: %d",
            len(_unindex_keys),
            len(set(_unindex_keys)),
        )
        logger.info(
            "Number of _unindex uid values:  %d, unique: %d",
            len(_unindex_values),
            len(set(_unindex_values)),
        )

        # Both index and unindex should have the same number of entries.
        # And keys and values should be unique.
        if not (
            len(_index_keys)
            == len(set(_index_keys))
            == len(_index_values)
            == len(set(_index_values))
            == len(_unindex_keys)
            == len(set(_unindex_keys))
            == len(_unindex_values)
            == len(set(_unindex_values))
        ):
            logger.error(
                "The numbers of keys and values in the UID index are NOT consistent."
            )
            return False

        logger.info("The numbers of keys and values in the UID index are consistent.")
        # Are the matching sets the same?
        if set(_index_keys) != set(_unindex_values) or set(_index_values) != set(
            _unindex_keys
        ):
            logger.error(
                "The sets are NOT the same. Inconsistencies found in UID index."
            )
            return False
        logger.info("No inconsistencies found in UID index.")
        return True

    def _commit(self, note=""):
        """Commit the current transaction with an optional note."""
        if EUPHORIE_DISABLE_INTERMEDIATE_COMMITS:
            return
        tx = transaction.get()
        if note:
            tx.note(note)
        tx.commit()
        logger.info("Committed transaction: %s", note)

    def __call__(self):
        """Upgrade step to fix inconsistencies in the UID index.

        I will add some inline comments about how long this took on a test
        site with about 420k items and a messed up UID index.
        """
        logger.info("Checking inconsistencies in UID index.")
        if self.is_uid_index_consistent():
            logger.info("The UID index is consistent. Nothing to do.")
            return

        # Gather paths for which we will create a new uuid.
        recreate = set()
        catalog = api.portal.get_tool("portal_catalog")
        index = catalog.Indexes["UID"]
        # Get the values once to speed up the loop below.  Without this, the
        # first full run took over an hour, gathering 220k paths to recreate.
        # Make it a set for faster 'in' checks.  This reduces the time insanely.
        index_index_values = set(index._index.values())
        logger.info("Checking _unindex items.")
        for docid, uid in index._unindex.items():
            if docid not in index_index_values:
                path = catalog.getpath(docid)
                logger.debug(
                    "Doc id %s is missing from _index values. UID %s, path %s",
                    docid,
                    uid,
                    path,
                )
                recreate.add(path)
                if len(recreate) % 10000 == 0:
                    logger.info("Found %d paths to recreate so far.", len(recreate))

        logger.info("Done checking _unindex items.")
        if not recreate:
            logger.info("No paths need to have their UID recreated.")
            return
        if recreate:
            logger.info(
                "We will recreate %d UIDs/UUIDs that are currently duplicate.",
                len(recreate),
            )
            logger.info(
                "You might need to manually fix some links.\n"
                "We have no way of knowing if a link should use "
                "resolveuid/old_uid or resolveuid/new_uid.\n"
                "Perhaps we could query the relation catalog to see "
                "which relations an item has."
            )

        # Now recreate UIDs for the gathered paths.
        # This took about 6 minutes on the test site.
        app = self.portal.getPhysicalRoot()
        for num, path in enumerate(recreate, 1):
            try:
                obj = app.unrestrictedTraverse(path)
            except KeyError:
                logger.info("Ignoring unreachable path to recreate UID: %s", path)
                continue
            # obj.UID() would return the UID of the parent in case
            # obj is a Discussion Item.
            old_uuid = IUUID(obj)
            # This might find an item by acquisition.
            # migration-law/migration-law/migration-law/research.htm
            # may actually be migration-law/research.htm
            actual_path = "/".join(obj.getPhysicalPath())
            if actual_path != path:
                logger.warning(
                    "Wanted to recreate UID for path %s, but this leads to "
                    "other path %s. Ignoring.",
                    path,
                    actual_path,
                )
                continue
            delattr(obj, ATTRIBUTE_NAME)
            # Call the event handler that adds a UUID:
            addAttributeUUID(obj, None)
            new_uuid = IUUID(obj)
            logger.debug("Changed UID from %s to %s for %s", old_uuid, new_uuid, path)
            if num % 10000 == 0:
                note = f"Created fresh UID for {num}/{len(recreate)} paths so far."
                logger.info(note)
                if num % 50000 == 0:
                    self._commit(note=note)

        note = f"Created fresh UID for all {len(recreate)} paths."
        logger.info(note)
        self._commit(note=note)

        # Update catalog metadata to reflect new UIDs.
        # On the test site this took about 13 minutes.
        note = "Updating catalog metadata for UIDs."
        logger.info(note)
        logger.info("This can take a long time...")
        update_catalog_metadata(self.portal, "UID")
        self._commit(note=note)

        # Now we need to clear and reindex the UID index.
        logger.info("Clearing UID index...")
        index.clear()
        # Reindexing took about 10 minutes on the test site.
        # So let's add a progress logger.
        logger.info("Reindexing UID index...")
        catalog._catalog.reindexIndex(
            "UID",
            getRequest(),
            pghandler=ZLogHandler(10000),
        )
        note = "Reindexed UID index."
        logger.info(note)
        self._commit(note=note)

        logger.info("The UID index should be fine again. Checking...")
        if not self.is_uid_index_consistent():
            raise ValueError(
                "After all fixes and reindexing, the UID index is still inconsistent."
            )
        logger.info("Yes, the UID index is consistent. All done.")



from bs4 import BeautifulSoup
from email import encoders as Encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr
from email.utils import formatdate
from logging import getLogger
from plone import api
from plone.base.utils import safe_bytes
from plone.base.utils import safe_text
from plone.memoize.view import memoize
from plone.rfc822.interfaces import IPrimaryFieldInfo
from Products.Five import BrowserView
from urllib.parse import quote

import markdown
import stoneagehtml


logger = getLogger(__name__)


class BaseEmail(BrowserView):
    """Helper view that renders an email, this is meant to be used as a base
    class."""

    # Use DumbWriters word wrapping to ensure that no text line
    # is longer than plain_text_maxcols characters.
    plain_text_maxcols = 72

    translatable_subject = ""

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def subject(self):
        return api.portal.translate(self.translatable_subject)

    @property
    @memoize
    def recipient(self):
        """You must override this to return any recipient you like or even None
        if you plan to provide a proper To, Cc, Bcc header using the headers
        method."""
        raise NotImplementedError

    @property
    def sender(self):
        from_address = api.portal.get_registry_record("plone.email_from_address")
        from_name = api.portal.get_registry_record("plone.email_from_name")
        return formataddr((from_name, from_address))

    @property
    def reply_to(self):
        """You must override this to add a reply-to to header to the email."""
        pass

    @property
    def headers(self):
        """You do not normally need to provide the From, Subject and To headers
        here because they will be setup by the send_email method.

        Anyway this is not forbidden. If you want to CC or BCC people,
        add an header here.
        """
        headers = {"Date": formatdate(localtime=True)}
        # Some mail servers don't allow a comma in the recipient name even if
        # it's quoted.
        headers["To"] = self.recipient.replace(",", "")
        reply_to = self.reply_to
        if reply_to:
            headers["Reply-to"] = reply_to
        return headers

    @property
    def html_email(self):
        """It is normally html generated by the index method and stripped to
        work for every MUA.

        You can override the index method by registering a template in
        some zcml or with a plain python method.
        """
        html = markdown.markdown(self.index(), extensions=["nl2br"])
        return safe_bytes(stoneagehtml.compactify(html))

    @property
    def text_email(self):
        """By default transforms the html_email in to a text version."""
        # Produce an approximate textual rendering of the HTML string,
        # unless you have been given a better version as an argument
        soup = BeautifulSoup(self.html_email, features="lxml")
        # kill all script and style elements
        for script in soup(["script", "style"]):
            script.extract()

        anchors = []
        for cnt, it in enumerate(soup(["a"])):
            it.replace_with(f"{it.text}[{cnt + 1}]")
            anchors.append(it["href"])

        text = "{}\n\n{}".format(
            safe_text(soup.get_text()),
            "\n".join(f"[{cnt + 1}] {it}" for cnt, it in enumerate(anchors)),
        )

        # break into lines and remove leading and trailing space on each
        lines = (line.strip() for line in text.splitlines())
        # break multi-headlines into a line each
        chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
        # drop blank lines
        text = "\n".join(chunk for chunk in chunks if chunk)
        return text

    @property
    def attachments(self):
        """List here your attachments, they will be translated in to email
        parts by the attachment2part method."""
        return []

    def attachment2part(self, attachment):
        """Transform an attachment in to a part.

        Attachment is expected to be an IPrimaryFieldInfo adaptable
        object. If you want to attach something else you might want to
        override this method or provide some adapter
        """
        primary_field_info = IPrimaryFieldInfo(attachment)
        field = primary_field_info.value
        try:
            mimetype = field.contentType
        except Exception:
            logger.exception("Cannot get mimetype for %r", attachment)
            mimetype = "application/octet-stream"

        maintype, subtype = mimetype.partition("/")[::2]
        if maintype == "text":
            part = MIMEText(field.data, _subtype=subtype, _charset="UTF-8")
        else:
            part = MIMEBase(maintype, subtype)
            part.set_payload(field.data)
            Encoders.encode_base64(part)
        part.add_header(
            "Content-Disposition",
            "attachment;filename*=UTF-8''%s" % quote(field.filename),
        )
        return part

    @property
    def body(self):
        """Create a mime-message that will render HTML in popular MUAs, text in
        better ones.

        Taken with thanks from ftw.notification.email.
        """
        # if we would like to include images in future, there should
        # probably be 'related' instead of 'mixed'
        msg = MIMEMultipart("mixed")
        for key, value in self.headers.items():
            msg[key] = value

        msg.preamble = "This is a multi-part message in MIME format."

        alternatives = MIMEMultipart("alternative")
        msg.attach(alternatives)
        alternatives.attach(MIMEText(self.text_email, "plain", _charset="UTF-8"))
        alternatives.attach(MIMEText(self.html_email, "html", _charset="UTF-8"))

        # add the attachments
        for attachment in self.attachments:
            part = self.attachment2part(attachment)
            msg.attach(part)
        return msg

    def send_email(self):
        body = self.body
        mh = api.portal.get_tool("MailHost")
        mh.send(body, mfrom=self.sender, subject=self.subject, charset="utf-8")

    def __call__(self):
        return self



from euphorie.deployment.upgrade.utils import alembic_upgrade_to
from ftw.upgrade import UpgradeStep


class AddTrainingQuestion(UpgradeStep):
    """Add Training Question."""

    def __call__(self):
        self.install_upgrade_profile()
        alembic_upgrade_to(self.target_version)



from AccessControl import getSecurityManager
from AccessControl import Unauthorized
from Acquisition import aq_inner
from euphorie.content import MessageFactory as _
from euphorie.content.sector import ISector
from euphorie.content.survey import ISurvey
from plone import api
from plonetheme.nuplone import MessageFactory as nu_
from plonetheme.nuplone.skin import sitemenu
from plonetheme.nuplone.utils import checkPermission
from plonetheme.nuplone.utils import getPortal
from Products.CMFCore.utils import getToolByName
from Products.membrane.interfaces.user import IMembraneUser


class Sitemenu(sitemenu.Sitemenu):
    @property
    def actions(self):
        """See plonetheme.nuplone.skin.sitemenu.py."""
        menu = super().actions or {}
        children = menu.get("children")
        if not children:
            return None

        submenu = self.menu_country_tools()
        if submenu:
            self.add_submenu(children, submenu)
        if children:
            return menu
        else:
            return None

    def menu_country_tools(self):
        context = aq_inner(self.context)

        # We try to traverse to the view. It would fail for the wrong context
        # (not an ICountry) or if permissions are not met.
        try:
            self.context.restrictedTraverse("@@country-tools")
        except (AttributeError, Unauthorized):
            return None

        menu = {"title": _("menu_admin", default="Admin")}
        menu["children"] = [
            {
                "title": _("menu_country_tools", default="Tools for this country"),
                "url": "%s/@@country-tools" % context.absolute_url(),
            }
        ]
        return menu

    @property
    def settings_url(self):
        user = getSecurityManager().getUser()
        if IMembraneUser.providedBy(user):
            mt = getToolByName(self.context, "membrane_tool")
            home = mt.getUserObject(user_id=user.getUserId())
            return "%s/@@edit" % home.absolute_url()
        else:
            home = getPortal(self.context)
            return "%s/@@settings" % home.absolute_url()

    def organise(self):
        menu = super().organise()
        if menu is not None:
            children = menu["children"]
        else:
            menu = {"title": nu_("menu_organise", default="Organise")}
            children = menu["children"] = []

        context_url = aq_inner(self.context).absolute_url()
        if ISurvey.providedBy(self.context) and checkPermission(self.context, "View"):
            children.append(
                {
                    "title": _("menu_export", default="XML export"),
                    "url": "%s/@@export" % context_url,
                }
            )
        if ISector.providedBy(self.context) and checkPermission(
            self.context, "Euphorie: Add new RIE Content"
        ):
            children.append(
                {
                    "title": _("menu_import", default="Import OiRA Tool"),
                    "url": "%s/@@upload" % context_url,
                }
            )

        if api.user.has_permission("Manage portal"):
            children.append(
                {
                    "title": _("Maintenance view"),
                    "url": f"{api.portal.get().absolute_url()}/@@admin-maintenance",
                }
            )

        if children:
            return menu
        else:
            return None



from AccessControl import Unauthorized
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from euphorie.client import model
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.model import Consultancy
from euphorie.client.model import Organisation
from euphorie.client.model import OrganisationMembership
from euphorie.client.model import SessionEvent
from euphorie.client.tests.utils import addSurvey
from euphorie.client.tests.utils import MockMailFixture
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from z3c.saconfig import Session
from zope.interface import alsoProvides

import json


class TestSessionValidation(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        with api.env.adopt_user("admin"):
            addSurvey(self.portal, BASIC_SURVEY)

        self.session = Session()

        self.owner = model.Account(
            loginname="valerie@labyrinth.social", password="secret"
        )
        self.session.add(self.owner)
        self.session.flush()
        self.session.add(
            Organisation(
                owner_id=self.owner.id,
            )
        )

        survey_session = model.SurveySession(
            id=1,
            title="Euphorie",
            zodb_path="nl/ict/software-development",
            account=self.owner,
        )
        self.session.add(survey_session)
        survey = self.portal.client.nl.ict["software-development"]
        session_id = "++session++%d" % survey_session.id
        self.traversed_session = survey.restrictedTraverse(session_id)

        self.consultant = model.Account(
            loginname="michel.moulin@example-consultancy.com",
            password="secret",
        )
        self.session.add(self.consultant)
        self.session.flush()

        self.session.add(
            OrganisationMembership(
                owner_id=self.owner.id,
                member_id=self.consultant.id,
                member_role="consultant",
            )
        )

        self.session.flush()
        alsoProvides(self.request, IClientSkinLayer)

    def test_available_consultants(self):
        with api.env.adopt_user(user=self.owner):
            with self._get_view(
                "panel-request-validation",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                self.assertIn(self.consultant, view.consultants)

    def test_request_validation_permission_denied(self):
        other_member = model.Account(
            loginname="siobhan@labyrinth.social",
            password="secret",
        )
        self.session.add(other_member)
        self.session.flush()

        self.session.add(
            OrganisationMembership(
                owner_id=self.owner.id,
                member_id=other_member.id,
                member_role="member",
            )
        )
        self.session.flush()

        with api.env.adopt_user(user=other_member):
            with self._get_view(
                "panel-request-validation",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                self.assertRaises(Unauthorized, view)

    def test_request_validation(self):
        mail_fixture = MockMailFixture()
        with api.env.adopt_user(user=self.owner):
            with self._get_view(
                "panel-request-validation",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                view.request.method = "POST"
                view.request.form = {"consultant": self.consultant.id}
                view()
                self.assertEqual(
                    self.traversed_session.session.consultancy.account,
                    self.consultant,
                )
                events = (
                    self.session.query(SessionEvent)
                    .filter(
                        SessionEvent.session_id == self.traversed_session.session.id
                    )
                    .order_by(SessionEvent.time.desc())
                    .all()
                )
                self.assertEqual(len(events), 1)
                self.assertEqual(events[0].action, "validation_requested")
                self.assertEqual(events[0].account_id, self.owner.id)
                self.assertLessEqual(
                    abs(events[0].time - datetime.utcnow()),
                    timedelta(seconds=5),
                )
                self.assertEqual(len(mail_fixture.storage), 1)
                self.assertEqual(
                    mail_fixture.storage[0][0][1],
                    "michel.moulin@example-consultancy.com",
                )

    def test_validate_permission_denied_to_other_member(self):
        other_member = model.Account(
            loginname="siobhan@labyrinth.social",
            password="secret",
        )
        self.session.add(other_member)
        self.session.flush()

        self.session.add(
            OrganisationMembership(
                owner_id=self.owner.id,
                member_id=other_member.id,
                member_role="member",
            )
        )
        self.session.flush()

        self.traversed_session.session.consultancy = Consultancy(
            account=self.consultant,
            session=self.traversed_session.session,
        )
        with api.env.adopt_user(user=other_member):
            with self._get_view(
                "panel-validate-risk-assessment",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                self.assertRaises(Unauthorized, view)

    def test_validate_permission_denied_to_other_consultant(self):
        other_consultant = model.Account(
            loginname="nancy.ann.ciancy@example-consultancy.com",
            password="secret",
        )
        self.session.add(other_consultant)
        self.session.flush()

        self.session.add(
            OrganisationMembership(
                owner_id=self.owner.id,
                member_id=other_consultant.id,
                member_role="consultant",
            )
        )

        self.session.flush()

        self.traversed_session.session.consultant = self.consultant
        with api.env.adopt_user(user=other_consultant):
            with self._get_view(
                "panel-validate-risk-assessment",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                self.assertRaises(Unauthorized, view)

    def test_validate_risk_assessment(self):
        second_admin = model.Account(
            loginname="jessica@labyrinth.social", password="secret"
        )
        self.session.add(second_admin)
        self.session.flush()

        self.session.add(
            OrganisationMembership(
                owner_id=self.owner.id,
                member_id=second_admin.id,
                member_role="admin",
            )
        )
        self.consultant.first_name = "Michel"
        self.consultant.last_name = "Moulin"

        mail_fixture = MockMailFixture()
        self.traversed_session.session.consultancy = Consultancy(
            account=self.consultant,
            session=self.traversed_session.session,
        )
        self.session.flush()
        with api.env.adopt_user(user=self.consultant):
            with self._get_view(
                "panel-validate-risk-assessment",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                view.request.method = "POST"
                view.request.form = {"approved": "1"}
                view()

        # consultant stays set
        self.assertEqual(
            self.traversed_session.session.consultancy.account,
            self.consultant,
        )
        self.assertTrue(self.traversed_session.session.is_validated)
        with api.env.adopt_user(user=self.owner):
            with self._get_view(
                "consultancy",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                self.assertEqual(
                    view.validated_info.consultant_email,
                    "michel.moulin@example-consultancy.com",
                )
                self.assertEqual(view.validated_info.consultant_name, "Michel Moulin")
                self.assertLessEqual(
                    abs(
                        view.validated_info.raw_time
                        - datetime.utcnow().replace(tzinfo=timezone.utc)
                    ),
                    timedelta(seconds=5),
                )

        self.assertTrue(self.traversed_session.session.is_locked)

        self.assertEqual(len(mail_fixture.storage), 2)
        recipients = {
            mail_fixture.storage[0][0][1],
            mail_fixture.storage[1][0][1],
        }
        self.assertSetEqual(
            recipients,
            {"valerie@labyrinth.social", "jessica@labyrinth.social"},
        )

    def test_delete_consultant(self):
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        self.traversed_session.session.consultancy = Consultancy(
            account=self.consultant,
            session=self.traversed_session.session,
        )
        event = SessionEvent(
            account_id=self.consultant.id,
            session_id=self.traversed_session.session.id,
            action="validated",
            time=now,
            note=json.dumps(
                {
                    "consultant_id": self.consultant.id,
                    "consultant_email": self.consultant.email,
                    "consultant_name": "Michel Moulin",
                }
            ),
        )
        self.session.add(event)

        self.session.delete(self.consultant)
        self.session.flush()

        self.assertTrue(self.traversed_session.session.is_validated)
        with api.env.adopt_user(user=self.owner):
            with self._get_view(
                "consultancy",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                self.assertEqual(
                    view.validated_info.consultant_email,
                    "michel.moulin@example-consultancy.com",
                )
                self.assertEqual(view.validated_info.consultant_name, "Michel Moulin")
            self.assertLessEqual(
                abs(view.validated_info.raw_time - now),
                timedelta(seconds=5),
            )

    def test_owner_invalidates_risk_assessment(self):
        self.traversed_session.session.consultancy = Consultancy(
            account=self.consultant,
            session=self.traversed_session.session,
        )
        event = SessionEvent(
            account_id=self.owner.id,
            session_id=self.traversed_session.session.id,
            action="validation_requested",
            time=(datetime.utcnow() - timedelta(days=2)).replace(tzinfo=timezone.utc),
        )
        self.session.add(event)
        event = SessionEvent(
            account_id=self.consultant.id,
            session_id=self.traversed_session.session.id,
            action="validated",
            time=(datetime.utcnow() - timedelta(days=1)).replace(tzinfo=timezone.utc),
            note=json.dumps(
                {
                    "consultant_id": self.consultant.id,
                    "consultant_email": self.consultant.email,
                    "consultant_name": "Michel Moulin",
                }
            ),
        )
        self.session.add(event)

        with api.env.adopt_user(user=self.owner):
            with self._get_view(
                "panel-invalidate-risk-assessment",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                view.request.method = "POST"
                view()

        self.assertFalse(self.traversed_session.session.is_locked)

    def test_invalidate_risk_assessment_permission_denied(self):
        self.traversed_session.session.consultancy = Consultancy(
            account=self.consultant,
            session=self.traversed_session.session,
        )
        event = SessionEvent(
            account_id=self.owner.id,
            session_id=self.traversed_session.session.id,
            action="validation_requested",
            time=(datetime.utcnow() - timedelta(days=2)).replace(tzinfo=timezone.utc),
        )
        self.session.add(event)
        event = SessionEvent(
            account_id=self.consultant.id,
            session_id=self.traversed_session.session.id,
            action="validated",
            time=(datetime.utcnow() - timedelta(days=1)).replace(tzinfo=timezone.utc),
            note=json.dumps(
                {
                    "consultant_id": self.consultant.id,
                    "consultant_email": self.consultant.email,
                    "consultant_name": "Michel Moulin",
                }
            ),
        )
        self.session.add(event)

        with api.env.adopt_user(user=self.consultant):
            with self._get_view(
                "panel-invalidate-risk-assessment",
                self.traversed_session,
                self.traversed_session.session,
            ) as view:
                view.request.method = "POST"
                self.assertRaises(Unauthorized, view)

        self.assertTrue(self.traversed_session.session.is_locked)



from euphorie.content import MessageFactory as _
from plone import api
from plone.base.utils import safe_text
from plone.memoize.view import memoize
from plone.memoize.view import memoize_contextless
from Products.Five import BrowserView


TYPES_MAP = {
    "euphorie.documentation": "Documentation",
    "euphorie.help": "Help",
    "euphorie.sector": _("Sector"),
    "euphorie.module": _("Module"),
    "euphorie.profilequestion": _("Profile question"),
    "euphorie.risk": _("Risk"),
    "euphorie.survey": _("OiRA Tool version"),
    "euphorie.surveygroup": _("OiRA Tool"),
    "euphorie.page": "Page",
}

SEARCHED_TYPES = list(TYPES_MAP.keys())


class Search(BrowserView):
    @property
    @memoize_contextless
    def did_search(self):
        qs = self.request.form.get("q", None)
        return qs is not None

    @property
    @memoize_contextless
    def results(self):
        qs = self.request.form.get("q", None)
        if not qs:
            return

        query = {"SearchableText": qs, "portal_type": SEARCHED_TYPES}
        ct = api.portal.get_tool("portal_catalog")
        return ct.searchResults(**query)


class ContextSearch(BrowserView):
    @property
    @memoize
    def did_search(self):
        qs = self.request.form.get("q", None)
        return qs is not None and qs.strip() != ""

    @property
    @memoize
    def results(self):
        qs = self.request.form.get("q", None)
        if not qs:
            return

        qs = f'"{safe_text(qs)}*"'
        path = "/".join(self.context.getPhysicalPath())
        query = {"SearchableText": qs, "portal_type": SEARCHED_TYPES, "path": path}

        ct = api.portal.get_tool("portal_catalog")
        brains = ct.searchResults(**query)

        return [
            {
                "url": brain.getURL(),
                "title": brain.Title,
                "typ": TYPES_MAP.get(brain.portal_type, "unknown"),
            }
            for brain in brains
        ]



from euphorie.content.behaviour.dirtytree import clearDirty
from euphorie.content.behaviour.dirtytree import isDirty
from euphorie.content.tests.utils import addSurvey
from euphorie.content.tests.utils import createSector
from euphorie.content.tests.utils import EMPTY_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent


class DirtyTreeTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()

    def create(self):
        sector = createSector(self.portal)
        return addSurvey(sector, EMPTY_SURVEY)

    def testStartClean(self):
        survey = self.create()
        self.assertEqual(isDirty(survey), False)

    def testAddObjectMakesDirty(self):
        survey = self.create()
        survey.invokeFactory("euphorie.module", "module")
        self.assertEqual(isDirty(survey), True)

    def testDeleteObjetMakesDiry(self):
        survey = self.create()
        survey.invokeFactory("euphorie.module", "module")
        clearDirty(survey)
        del survey["module"]
        self.assertEqual(isDirty(survey), True)

    def testModifyObjectMakesDirty(self):
        survey = self.create()
        survey.invokeFactory("euphorie.module", "module")
        clearDirty(survey)
        notify(ObjectModifiedEvent(survey["module"]))
        self.assertEqual(isDirty(survey), True)


class RichDescriptionTests(EuphorieIntegrationTestCase):
    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createModule(self):
        with api.env.adopt_user("admin"):
            self.country = self.portal.sectors.nl
            self.sector = self._create(self.country, "euphorie.sector", "sector")
            self.surveygroup = self._create(
                self.sector, "euphorie.surveygroup", "group"
            )
            self.survey = self._create(self.surveygroup, "euphorie.survey", "survey")
            self.module = self._create(self.survey, "euphorie.module", "module")
        return self.module

    def testNoMarkup(self):
        module = self.createModule()
        module.description = "Raw text"
        module.indexObject()
        brain = self.portal.portal_catalog(portal_type="euphorie.module")[0]
        self.assertEqual(brain.Description, "Raw text")

    def testStrayBracket(self):
        module = self.createModule()
        module.description = "Test <em>me</em> >"
        module.indexObject()
        brain = self.portal.portal_catalog(portal_type="euphorie.module")[0]
        self.assertEqual(brain.Description, "Test me >")

    def testNone(self):
        module = self.createModule()
        module.description = None
        module.indexObject()
        brain = self.portal.portal_catalog(portal_type="euphorie.module")[0]
        self.assertEqual(brain.Description, None)



"""Add the columns first_name and last_name to the account table.

Revision ID: 20220620103222
Revises: 20220318135716
Create Date: 2022-06-20 10:33:43.713596
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20220620103222"
down_revision = "20220318135716"
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("account", "first_name"):
        op.add_column("account", sa.Column("first_name", sa.Unicode(), nullable=True))
    if not has_column("account", "last_name"):
        op.add_column("account", sa.Column("last_name", sa.Unicode(), nullable=True))


def downgrade():
    op.drop_column("account", "last_name")
    op.drop_column("account", "first_name")



from AccessControl import getSecurityManager
from Acquisition import aq_inner
from Acquisition import aq_parent
from euphorie.client.browser.webhelpers import WebHelpers
from euphorie.client.model import ActionPlan
from euphorie.client.model import Session
from euphorie.client.model import SurveySession
from euphorie.client.model import SurveyTreeItem
from euphorie.content.countrymanager import ICountryManager
from euphorie.content.risk import EnsureInterface
from euphorie.content.risk import IRisk
from plone import api
from plone.dexterity.interfaces import IDexterityContainer
from plone.memoize.instance import memoize
from plone.memoize.view import memoize_contextless
from plone.protect.interfaces import IDisableCSRFProtection
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.membrane.interfaces.user import IMembraneUser
from sqlalchemy import and_
from sqlalchemy import sql
from time import time
from zope.deprecation import deprecate
from zope.interface import alsoProvides

import logging


log = logging.getLogger(__name__)


class Frontpage(BrowserView):
    def __call__(self):
        user = getSecurityManager().getUser()
        if IMembraneUser.providedBy(user):
            mt = getToolByName(self.context, "membrane_tool")
            obj = mt.getUserObject(user_id=user.getUserId())
            if ICountryManager.providedBy(obj):
                return self.request.response.redirect(aq_parent(obj).absolute_url())
            return self.request.response.redirect(obj.absolute_url())
        portal = aq_inner(self.context)
        return self.request.response.redirect("%s/sectors/" % portal.absolute_url())


class EuphorieRefreshResourcesTimestamp(BrowserView):
    @property
    @memoize_contextless
    def resources_timestamp(self):
        return api.portal.get_registry_record(
            "euphorie.deployment.resources_timestamp", default=""
        )

    def refresh_timestamp(self):
        alsoProvides(self.request, IDisableCSRFProtection)
        api.portal.set_registry_record(
            "euphorie.deployment.resources_timestamp", str(int(time()))
        )

    def __call__(self):
        """Refresh the registry record that adds a timestamp to the resources
        urls."""
        self.refresh_timestamp()
        return "OK"


class GetEuphorieResourcesTimestamp(EuphorieRefreshResourcesTimestamp):
    def __call__(self):
        """Get the resource timestamp."""
        return self.resources_timestamp


class ManageEnsureInterface(BrowserView):
    def set_evaluation_method_interfaces(self):
        def walk(node):
            for idx, sub_node in node.ZopeFind(node, search_sub=0):
                if IRisk.providedBy(sub_node):
                    yield sub_node
                if IDexterityContainer.providedBy(sub_node):
                    yield from walk(sub_node)

        count = 0
        walker = walk(self.context)
        for risk in walker:
            log.debug("Ensure interface for %s", "/".join(risk.getPhysicalPath()))
            EnsureInterface(risk)
            count += 1
        log.info("Ensured interface for %d risks", count)
        return count

    def __call__(self):
        """Iterate over all risks in the current context and ensure that they
        have the correct interface."""
        count = self.set_evaluation_method_interfaces()
        return "Handled %d risks" % count


class UpdateCompletionPercentage(WebHelpers):
    """Utility view to fill in missing values for completion_percentage."""

    flush_threshold = 50

    def log(self, entry):
        log.info(entry)
        self._log = "\n".join((getattr(self, "_log", ""), entry))

    def get_log(self):
        return getattr(self, "_log", "")

    def next_b_start(self):
        if "overwrite" in self.request.form:
            return self.request.get("b_start", 0) + self.request.get("b_size", 1000)
        else:
            return 0

    @deprecate(
        "Not needed anymore because the completion_percentage is now calculated. "
        "Deprecated in version 14.1.4.dev0"
    )
    def __call__(self):
        return self.index()


class RepairSolutionId(BrowserView):
    @memoize
    def get_tool(self, tool_path):
        return self.client.restrictedTraverse(str(tool_path))

    @memoize
    def get_risk(self, tool, risk_path):
        return tool.restrictedTraverse(str(risk_path))

    def __call__(self):
        site = api.portal.get()
        self.client = getattr(site, "client")
        query = (
            Session.query(ActionPlan, SurveyTreeItem, SurveySession)
            .filter(
                sql.and_(
                    ActionPlan.plan_type == "in_place_standard",
                    ActionPlan.solution_id == None,  # noqa: E711
                )
            )
            .join(SurveyTreeItem, ActionPlan.risk_id == SurveyTreeItem.id)
            .join(SurveySession, SurveyTreeItem.session_id == SurveySession.id)
            .order_by(SurveySession.zodb_path)
        )
        count = 0
        ret = "We have a total of %d measures\n" % query.count()

        for action_plan, risk, session in query.all():
            tool = self.get_tool(session.zodb_path)
            risk = self.get_risk(tool, risk.zodb_path)
            for solution in risk._solutions:
                if (
                    solution.action
                    and action_plan.action
                    and solution.action.strip() == action_plan.action.strip()
                ):
                    action_plan.solution_id = str(solution.id)
                    count += 1
                    break
        ret += "Updated %d measures" % count
        return ret


class FixOmegaPaths(BrowserView):
    def __call__(self):
        session = Session()

        custom_risks = session.query(SurveyTreeItem).filter(
            and_(
                SurveyTreeItem.zodb_path.notlike("custom-risks%"),
                SurveyTreeItem.zodb_path.like("%custom-risks%"),
                SurveyTreeItem.type == "risk",
            )
        )

        count = 0

        for risk in custom_risks:
            i = 1
            num = risk.zodb_path.split("/")[-1]
            while i < 100:
                new_zodb_path = f"custom-risks/{num}"
                conflict = session.query(SurveyTreeItem).filter(
                    and_(
                        SurveyTreeItem.session_id == risk.session_id,
                        SurveyTreeItem.zodb_path == new_zodb_path,
                    )
                )
                if conflict.count():
                    num = num = str(int(num) + i)
                else:
                    risk.zodb_path = new_zodb_path
                    count += 1
                    break
                i += 1
        return "Fixed %d omega risks" % count



from euphorie.client.country import ClientCountry
from euphorie.testing import EuphorieIntegrationTestCase
from zope.annotation import IAttributeAnnotatable
from zope.interface import implementer


class ViewTests(EuphorieIntegrationTestCase):
    def addSurvey(self, country, language, title):
        from euphorie.client.sector import ClientSector
        from euphorie.content.survey import Survey

        if "sector" not in country:
            sector = ClientSector(id="sector", title="Test sector")
            country._setOb("sector", sector)
        else:
            sector = country["sector"]
        survey = Survey()
        survey.title = title
        survey.id = language
        survey.language = language
        sector._setOb("language", survey)
        return survey

    def surveys(self, country, language):
        @implementer(IAttributeAnnotatable)
        class Request:
            environ = {"REQUEST_METHOD": "GET"}
            form = {}

            def __init__(self, language):
                self.language = language

            def __getattr__(self, key):
                if key in ["request", "locale", "id"]:
                    return self
                raise AttributeError(key)

        with self._get_view("view", country) as view:
            view.request = Request(language)
            view._updateSurveys()
            return view.surveys

    def test_surveys_NoSurveys(self):
        country = ClientCountry()
        self.assertEqual(self.surveys(country, "en"), [])

    def test_surveys_WrongLanguage(self):
        country = ClientCountry()
        self.addSurvey(country, "en", "Test survey")
        self.assertEqual(self.surveys(country, "nl"), [])

    def test_surveys_SameLanguage(self):
        country = ClientCountry()
        self.addSurvey(country, "en", "Test survey")
        self.assertEqual(
            self.surveys(country, "en"), [{"id": "sector/en", "title": "Test survey"}]
        )

    def test_surveys_SurveyHasDialect(self):
        country = ClientCountry()
        self.addSurvey(country, "en-GB", "Test survey")
        self.assertEqual(
            self.surveys(country, "en"),
            [{"id": "sector/en-GB", "title": "Test survey"}],
        )

    def test_surveys_SkipPreview(self):
        country = ClientCountry()
        survey = self.addSurvey(country, "en", "Test survey")
        survey.preview = True
        self.assertEqual(self.surveys(country, "en"), [])



from Acquisition import aq_chain
from Acquisition import aq_inner
from euphorie.content import MessageFactory as _
from euphorie.content.country import ICountry
from plone import api
from plone.memoize.view import memoize_contextless
from plonetheme.nuplone.tiles.tabs import TabsTile

import re


class SiteRootTabsTile(TabsTile):
    current_map = [
        (re.compile(r"/sectors/[a-z]+/*@@manage-users"), "usermgmt"),
        (re.compile(r"/sectors/[a-z]+/help"), "help"),
        (re.compile(r"/sectors"), "sectors"),
        (re.compile(r"/documents"), "documents"),
    ]

    @property
    @memoize_contextless
    def portal(self):
        return api.portal.get()

    @property
    @memoize_contextless
    def user(self):
        return api.user.get_current()

    def get_current_country(self):
        for obj in aq_chain(aq_inner(self.context)):
            if ICountry.providedBy(obj):
                return obj

    def is_country_manager(self):
        """Check if we are in the context of a country and if we have enough
        permissions to manage it."""
        country = self.get_current_country()
        if not country:
            return False
        if api.user.has_permission("Euphorie: Manage country", obj=country):
            return True

    def get_current_url(self):
        currentUrl = self.request.getURL()[len(self.portal.absolute_url()) :]
        for test, id in self.current_map:
            if test.match(currentUrl):
                return id

    def update(self):
        current = self.get_current_url()
        self.tabs = [
            {
                "id": "sectors",
                "title": _("nav_surveys", default="OiRA Tools"),
                "url": self.portal.sectors.absolute_url(),
                "class": "current" if current == "sectors" else None,
            }
        ]
        is_country_manager = self.is_country_manager()
        country = self.get_current_country()
        country_url = country and country.absolute_url() or ""
        if is_country_manager:
            self.tabs.append(
                {
                    "id": "usermgmt",
                    "title": _("nav_usermanagement", default="User management"),
                    "url": "%s/@@manage-users" % country_url,
                    "class": "current" if current == "usermgmt" else None,
                }
            )

        if api.user.has_permission("Manage portal"):
            self.tabs.append(
                {
                    "id": "documents",
                    "title": _("nav_documents", default="Documents"),
                    "url": self.portal.documents.absolute_url(),
                    "class": "current" if current == "documents" else None,
                }
            )

        if country:
            self.tabs.append(
                {
                    "id": "help",
                    "title": _("nav_help", default="Help"),
                    "url": "%s/help" % country_url,
                    "class": "current" if current == "help" else None,
                }
            )

        self.home_url = self.portal.absolute_url()



from Acquisition import aq_inner
from anytree import NodeMixin
from anytree.node.util import _repr
from euphorie import MessageFactory as _
from euphorie.client import utils
from euphorie.client.browser.webhelpers import WebHelpers
from euphorie.client.country import IClientCountry
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.model import get_current_account
from euphorie.client.sector import IClientSector
from euphorie.content.survey import ISurvey
from euphorie.content.utils import getRegionTitle
from logging import getLogger
from plone import api
from plone.base.utils import safe_text
from plone.memoize.view import memoize
from plone.memoize.view import memoize_contextless
from Products.Five import BrowserView
from typing import cast
from z3c.saconfig import Session
from zExceptions import Unauthorized
from zope.deprecation import deprecate
from zope.interface import directlyProvides

import json


logger = getLogger(__name__)


def capitalize(text):
    if text:
        return f"{text[0].upper()}{text[1:]}"


class Node(NodeMixin):
    def __init__(self, context, parent=None, **kwargs):
        self.__dict__.update(kwargs)
        self.context = context
        self.parent = parent

    @property
    def groups(self):
        """Assume childrens are groups and return them sorted by title."""
        return sorted(self.children, key=lambda x: x.title)

    @property
    def sessions(self):
        """Assume childrens are sessions and return them sorted by reversed
        modification date."""
        return sorted(self.children, key=lambda x: x.context.modified, reverse=True)

    @property
    def survey_templates(self):
        """Return all children that are survey_templates, sorted by title."""
        return sorted(
            (item for item in self.children if item.type == "survey_template"),
            key=lambda x: x.title.lower(),
        )

    @property
    def categories(self):
        """Return all children that are categories, sorted by title."""
        return sorted(
            (item for item in self.children if item.type == "category"),
            key=lambda x: x.title,
        )

    def __repr__(self):
        args = [
            "%r"
            % self.separator.join([""] + [repr(node.context) for node in self.path])
        ]
        return _repr(self, args=args, nameblacklist=["context"])


class SurveyTemplatesMixin:
    # Here, we assemble the list of available tools for starting a new session

    @property
    @memoize
    def survey_templates_root(self):
        return Node(None, title="", type="root")

    @memoize
    def get_survey_category_node(self, category):
        if category is None:
            # Everything is grouped under the survey_templates_root node
            return self.survey_templates_root
        return Node(
            category,
            parent=self.get_survey_category_node(None),
            title=category,
            type="category",
        )

    @memoize
    def get_survey_template_node(self, survey_item):
        category, survey, id = survey_item
        return Node(
            survey,
            parent=self.get_survey_category_node(category),
            title=survey.title,
            type="survey_template",
            id=id,
        )

    @memoize
    def get_survey_templates_tree_root(self):
        survey_items = self.get_survey_templates()
        list(map(self.get_survey_template_node, survey_items))
        return self.survey_templates_root

    @memoize
    def get_survey_templates(self):
        # this is a list of tuples of the form
        # (category name, survey object, survey id)
        survey_items = []
        language = self.request.locale.id.language or ""
        for sector in aq_inner(self.context).values():
            if not IClientSector.providedBy(sector):
                continue

            for survey in sector.values():
                if not ISurvey.providedBy(survey):
                    continue
                if getattr(survey, "preview", False):
                    continue
                if (
                    survey.language
                    and survey.language != language
                    and not survey.language.strip().startswith(language)
                ):
                    continue
                if getattr(survey, "obsolete", False):
                    continue
                categories = getattr(survey, "tool_category", None)
                id = f"{sector.id}/{survey.id}"
                if not isinstance(categories, list):
                    categories = [categories]
                if not categories:
                    categories = ["None"]
                for category in categories:
                    survey_items.append((category, survey, id))
        return sorted(survey_items, key=lambda x: (x[0] or "None", x[1].title))


class SessionsView(BrowserView, SurveyTemplatesMixin):
    variation_class = "variation-dashboard"
    form_action_name = "dashboard-switcher"

    @property
    @memoize
    def webhelpers(self) -> WebHelpers:
        return cast(
            WebHelpers, api.content.get_view("webhelpers", self.context, self.request)
        )

    @property
    @memoize
    def survey_session_model(self):
        return self.webhelpers.survey_session_model

    @property
    def _portlet_names(self):
        names = ["portlet-my-ras", "portlet-available-tools"]
        if self.webhelpers.use_training_module:
            names.append("portlet-my-trainings")
        return names

    @property
    @memoize
    def portlets(self):
        return [
            api.content.get_view(name, self.context, self.request)
            for name in self._portlet_names
        ]

    @property
    @memoize
    def my_context(self):
        if IClientCountry.providedBy(self.context):
            return "country"
        elif ISurvey.providedBy(self.context):
            return "survey"

    @property
    @memoize_contextless
    def portal(self):
        return api.portal.get()

    @property
    @memoize_contextless
    def account(self):
        """The currently authenticated account."""
        return get_current_account()

    def _updateSurveys(self):
        self.surveys = []
        self.obsolete_surveys = []

        language = self.request.locale.id.language or ""
        for sector in aq_inner(self.context).values():
            if not IClientSector.providedBy(sector):
                continue

            for survey in sector.values():
                if not ISurvey.providedBy(survey):
                    continue
                if getattr(survey, "preview", False):
                    continue
                if (
                    survey.language
                    and survey.language != language
                    and not survey.language.strip().startswith(language)
                ):
                    continue
                info = {"id": f"{sector.id}/{survey.id}", "title": survey.title}
                if getattr(survey, "obsolete", False):
                    # getattr needed for surveys which were published before
                    # the obsolete flag added.
                    self.obsolete_surveys.append(info)
                else:
                    self.surveys.append(info)
        self.surveys.sort(key=lambda s: s["title"])
        self.obsolete_surveys.sort(key=lambda s: s["title"])

    def _NewSurvey(self, info, account=None):
        """Utility method to start a new survey session."""
        context = aq_inner(self.context)
        survey = info.get("survey")
        survey = context.restrictedTraverse(survey)
        if not ISurvey.providedBy(survey):
            logger.error("Tried to start invalid survey %r", info.get("survey"))
            # Things are sufficiently messed up at this point that rendering
            # breaks, so trigger a redirect to the same URL again.
            self.request.response.redirect(context.absolute_url())
            return

        title = info.get("title", "").strip()

        survey_view = api.content.get_view("index_html", survey, self.request)
        survey_session = survey_view.create_survey_session(title, account)
        self.request.response.redirect(
            f"{survey_session.absolute_url()}/@@start?initial_view=1&new_session=1"
        )

    def _ContinueSurvey(self, info):
        """Utility method to continue an existing session."""
        session = Session.query(self.survey_session_model).get(info["session"])
        extra = ""
        if info.get("new_clone", None):
            extra = "&new_clone=1"
        self.request.response.redirect(
            f"{session.absolute_url()}/@@resume?initial_view=1{extra}"
        )

    def _CloneSurvey(self, info):
        session = Session.query(self.survey_session_model).get(info["session"])
        clone_view = api.content.get_view(
            "clone-session", session.traversed_session, self.request
        )
        clone_view.clone()

    def tool_byline(self):
        title = api.portal.translate(
            _("title_tool", default="OiRA - Online interactive Risk Assessment")
        )
        return title.split("-")[-1].strip()

    def set_language(self):
        utils.setLanguage(
            self.request, self.context, getattr(self.context, "language", None)
        )

    def __call__(self):
        if not self.account:
            raise Unauthorized()

        self.set_language()
        reply = self.request.form
        action = reply.get("action")
        if action == "new":
            if self.my_context == "survey":
                reply["survey"] = "/".join(self.context.getPhysicalPath())
            return self._NewSurvey(reply)
        elif action == "continue":
            return self._ContinueSurvey(reply)
        if action == "clone":
            return self._CloneSurvey(reply)
        self._updateSurveys()
        return super().__call__()


class SessionBrowserNavigator(BrowserView):
    """Logic to build the navigator for the sessions."""

    no_splash = True

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def group_model(self):
        return self.webhelpers.group_model

    @property
    @memoize
    def survey_session_model(self):
        return self.webhelpers.survey_session_model

    @property
    @memoize
    def groupid(self):
        return self.request.get("groupid")

    @memoize
    def get_root_group(self, groupid=None):
        """Return the group that is the root of the navigation tree."""
        if not groupid:
            groupid = self.groupid
        if not groupid:
            return
        base_query = Session.query(self.group_model).order_by(
            self.group_model.short_name
        )
        return base_query.filter(self.group_model.group_id == groupid).one_or_none()

    @property
    @memoize
    def searchable_text(self):
        """Return the text we need to search in postgres already surrounded
        with '%' Allow a minimum size of 3 characters to reduce the load."""
        searchable_text = self.request.get("SearchableText")
        if not isinstance(searchable_text, str):
            return ""
        if len(searchable_text) < 3:
            return ""
        return f"%{safe_text(searchable_text)}%"

    @memoize
    def leaf_groups(self, groupid=None):
        """Nothing to do in main OiRA - to be filled in customer-specific
        packages.
        """
        return []

    @memoize
    def leaf_sessions(self):
        """The sessions we want to display in the navigation."""
        query = self.webhelpers.get_sessions_query(
            context=self.webhelpers.country_obj, searchable_text=self.searchable_text
        )
        return query.all()

    def has_content(self):
        """Checks if we have something meaningfull to display."""
        if len(self.leaf_groups()):
            return True
        if len(self.leaf_sessions()):
            return True
        return False


class Assessments(BrowserView):
    query_parameter = "SearchableText"

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    def form_action_name(self):
        return self.__name__

    @property
    def template(self):
        return self.index

    @property
    @memoize
    def organisations(self):
        """Return the organisations the current user is member of."""
        organisation_view = api.content.get_view(
            "organisation", self.context, self.request
        )
        return organisation_view.organisations

    @property
    def organisation_options(self):
        """Return the organisations of the current user
        for displaying in a select box."""
        selected_organisation = self.selected_organisation
        return [
            {
                "label": organisation.title,
                "value": organisation.owner_id,
                "selected": (
                    "selected" if organisation == selected_organisation else None
                ),
            }
            for organisation in self.organisations
        ]

    def is_filter_active(self):
        """True if any filters in the request parameters are different from the
        defaults."""
        return (
            self.request.get("organisation_owner_id")
            or self.request.get("sort_on", "modified") != "modified"
        )

    @property
    def selected_organisation(self):
        """Return the organisation selected in the request
        if it matches an organisation of the currently authenticated users.
        """
        organisation_value = self.request.get("organisation_owner_id", "")
        try:
            organisation_value = int(organisation_value)
        except ValueError:
            return
        for organisation in self.organisations:
            if organisation.owner_id == organisation_value:
                return organisation

    @property
    @memoize
    def sessions(self):
        searchable_text = self.request.get(self.query_parameter, None)
        if searchable_text and "%" not in searchable_text:
            searchable_text = f"%{searchable_text}%"
        sort_on_value = self.request.get("sort_on", "modified")
        if sort_on_value == "alphabetical":
            order_by = self.webhelpers.survey_session_model.title
        else:
            order_by = False

        if self.selected_organisation:
            filter_by_account = self.selected_organisation.owner_id
        else:
            # `True` means filter for current account.
            filter_by_account = True

        return self.webhelpers.get_sessions_query(
            context=self.context,
            searchable_text=searchable_text,
            include_archived=True,
            filter_by_account=filter_by_account,
            order_by=order_by,
        )

    def get_archived_label(self, session):
        if not session.is_archived:
            return ""
        return api.portal.translate(_("Archived"))


class AssessmentsJson(Assessments):
    query_parameter = "q"

    def get_title(self, session):
        title = session.title
        if not title:
            title_missing = api.portal.translate(
                _("label_missing_title", default="Title is missing")
            )
            title = f"[{title_missing}]"
        return title

    def __call__(self):
        self.request.response.setHeader("Content-Type", "application/json")
        return json.dumps(
            [
                {"id": session.id, "text": self.get_title(session)}
                for session in self.sessions
            ]
        )


class Surveys(BrowserView, SurveyTemplatesMixin):
    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def form_action_name(self):
        return self.__name__

    def get_filters(self):
        filters = {}
        get = self.request.form.get

        client_path = "/".join(self.request.client.getPhysicalPath())
        path = "/".join((client_path, self.country))
        sector = get("sector")
        if sector:
            path = "/".join((path, sector))
        filters["path"] = path

        language = get("Language")
        if language:
            filters["Language"] = language

        sort_on = get("sort_on", "created")
        filters["sort_on"] = sort_on
        if sort_on == "created":
            filters["sort_order"] = "reverse"

        searchable_text = get("SearchableText")
        if searchable_text:
            filters["SearchableText"] = searchable_text

        return filters

    @property
    @memoize
    def country(self):
        return self.request.form.get("country", self.context.getId())

    @property
    @memoize
    def countries(self):
        return sorted(
            [
                {
                    "id": country.getId(),
                    "Title": getRegionTitle(self.request, country.getId()),
                }
                for country in self.request.client.values()
                if (IClientCountry.providedBy(country) and len(country.objectIds()))
            ],
            key=lambda co: "0" if co["id"] == "eu" else co["Title"],
        )

    @property
    @memoize
    def sectors(self):
        country_obj = self.request.client.restrictedTraverse(self.country)
        return [
            sector
            for sector in aq_inner(country_obj).values()
            if IClientSector.providedBy(sector)
        ]

    @property
    @memoize
    def tools(self):
        filters = self.get_filters()
        tools = [
            (None, survey.getObject(), survey.getId)
            for survey in api.content.find(
                object_provides="euphorie.content.survey.ISurvey", **filters
            )
        ]
        # We must filter out tools marked as obsolete or preview
        return [
            record
            for record in tools
            if (
                not getattr(record[1], "preview", None)
                and not getattr(record[1], "obsolete", None)
            )
        ]

    @property
    @memoize
    def languages(self):
        return sorted(
            [
                {"code": code, "name": self.webhelpers.getNameForLanguageCode(code)}
                for code in api.portal.get_tool("portal_catalog").uniqueValuesFor(
                    "Language"
                )
                if code and code != "None"
            ],
            key=lambda lang: lang["code"],
        )

    def __call__(self):
        utils.setLanguage(
            self.request, self.context, getattr(self.context, "language", None)
        )
        return super().__call__()


class PortletBase(BrowserView):
    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    def filter_survey(self, survey):
        if getattr(survey, "preview", False):
            return False
        if getattr(survey, "obsolete", False):
            return False
        if not survey.language:
            return True
        language = self.request.locale.id.language or ""
        if survey.language == language:
            return True
        if survey.language.strip().startswith(language):
            return True
        return False

    @property
    @memoize
    def surveys(self):
        surveys = set()
        sectors = self.context.listFolderContents(
            {"portal_type": "euphorie.clientsector"}
        )
        for sector in sectors:
            surveys.update(
                filter(
                    self.filter_survey,
                    sector.listFolderContents({"portal_type": "euphorie.survey"}),
                )
            )
        return sorted(surveys, key=lambda survey: survey.title)


class MyRAsPortlet(PortletBase):
    element_id = "portlet-my-risk-assessments"

    @property
    def columns(self):
        if self.surveys:
            return "2"
        return "3"

    @property
    def assessments_list_macro(self):
        """Get the listing macro from the assessments template."""
        request = self.request.clone()
        # This is needed, since the view might be subclassed and therefore
        # associated with a different skin-layer.
        directlyProvides(request, IClientSkinLayer)
        view = api.content.get_view("assessments", self.context, request)
        return view.template.macros["assessments_list"]

    @property
    @memoize_contextless
    def hide_archived(self):
        """By default we hide the archived session and we have a checkbox that
        shows with a sibling hide_archived_marker input field."""
        if self.request.get("hide_archived_marker"):
            if not self.request.get("hide_archived"):
                return False
        return True

    @property
    @memoize
    def organisation_view(self):
        """Get the organisation view to reuse its methods."""
        return api.content.get_view("organisation", self.context, self.request)

    @property
    @memoize
    @deprecate(
        "Replace with self.sessions_by_organisation (which is a dict), "
        "deprecated in version 14.2"
    )
    def sessions(self):
        """We want the archived sessions."""
        return (
            self.webhelpers.get_sessions_query(
                context=self.context, include_archived=not self.hide_archived
            )
            .limit(5)
            .all()
        )

    @property
    @memoize
    def sessions_by_organisation(self):
        """Return the sessions grouped by organisation."""
        base_query = self.webhelpers.get_sessions_query(
            context=self.context, include_archived=not self.hide_archived
        )

        account_id_column = self.webhelpers.survey_session_model.account_id

        sessions_by_organisation = []
        known_owner_ids = set()
        for organisation in self.organisation_view.organisations:
            known_owner_ids.add(organisation.owner_id)
            sessions = (
                base_query.filter(account_id_column == organisation.owner_id)
                .limit(5)
                .all()
            )
            if sessions:
                sessions_by_organisation.append((organisation, sessions))

        account = self.webhelpers.get_current_account()
        if not account.organisation:
            sessions = (
                base_query.filter(account_id_column.notin_(known_owner_ids))
                .limit(5)
                .all()
            )
            if sessions:
                sessions_by_organisation.insert(0, (None, sessions))

        return dict(sessions_by_organisation)

    @property
    def label_start_session(self):
        label = api.portal.translate(
            _("link_start_session", default="Start a new risk assessment")
        )
        return capitalize(label)


class AvailableToolsPortlet(PortletBase, SurveyTemplatesMixin):
    element_id = "portlet-available-tools"



from plonetheme.nuplone.z3cform.interfaces import INuPloneFormLayer
from zope.interface import Attribute
from zope.interface import implementer
from zope.interface import Interface
from zope.interface.interfaces import ObjectEvent
from zope.publisher.interfaces.browser import IDefaultBrowserLayer


INTERVAL_DAILY = "daily"
INTERVAL_HOURLY = "hourly"
INTERVAL_MINUTELY = "minutely"
INTERVAL_MANUAL = "manual"


class IClientSkinLayer(IDefaultBrowserLayer, INuPloneFormLayer):
    """Zope skin layer for the online client."""


class ICustomRisksModifiedEvent(Interface):
    """Custom risks were modified."""


@implementer(ICustomRisksModifiedEvent)
class CustomRisksModifiedEvent(ObjectEvent):
    """Custom risks were modified."""


class INotificationCategory(Interface):
    """A notification category adapter."""

    id = Attribute("Id")  # The id of the category
    title = Attribute("Title")  # The title of the category
    description = Attribute("Description")  # A description of the category

    # The interval, this notification should be checked.
    # One of "daily", "hourly", "minutely"
    interval = Attribute("Interval")

    # Return True if this category is available for the current user.
    available = Attribute("Available")

    def notify():
        """Send a notification."""



"""Add SessionEvent table

Revision ID: 20230504061955
Revises: 20230502091628
Create Date: 2023-05-04 06:20:23.137551

"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_table

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20230504061955"
down_revision = "20230502091628"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("session_event"):
        op.create_table(
            "session_event",
            sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
            sa.Column("time", sa.DateTime(), nullable=False),
            sa.Column("account_id", sa.Integer(), nullable=True),
            sa.Column("session_id", sa.Integer(), nullable=False),
            sa.Column("action", sa.Unicode(length=32), nullable=True),
            sa.Column("note", sa.Unicode(), nullable=True),
            sa.ForeignKeyConstraint(["account_id"], ["account.id"], onupdate="CASCADE"),
            sa.ForeignKeyConstraint(
                ["session_id"], ["session.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )


def downgrade():
    if has_table("session_event"):
        op.drop_table("session_event")



from euphorie.testing import EuphorieFunctionalTestCase
from euphorie.testing import EuphorieIntegrationTestCase
from plone.protect.authenticator import createToken
from z3c.saconfig import Session
from zExceptions import Forbidden
from zope.interface import alsoProvides

import transaction


class BaseProtectTestCase(EuphorieIntegrationTestCase):
    """setUp and methods used by the integration and functional test case."""

    def setUp(self):
        from euphorie.content.protect import EuphorieProtectTransform

        super().setUp()
        # The transform wants a 'published' and a 'request'.
        # 'published' can be a view.  But really it is ignored, so let's take
        # the portal.
        self.transform = EuphorieProtectTransform(self.portal, self.request)
        self.session = Session()
        self._latest_group_id = 0
        self.group_id = None

    def _createItem(self):
        """Create an SQL item.

        At first I tried a SurveyTreeItem, but eventually ran into problems
        because this would then automagically create a related Risk or Module.
        A session flush in test_edit_with_flush in the functional test case
        would fail with:

            sqlalchemy.orm.exc.ObjectDeletedError: Instance '<Risk at ...>'
            has been deleted, or its row is otherwise not present.

        This error seems unrelated to what we are trying to test here.
        So I try a simpler model.
        """
        from euphorie.client.model import Group

        # There are several constraints we need to satisfy.
        self._latest_group_id += 1
        group_id = str(self._latest_group_id)
        return Group(group_id=group_id)

    def _getItem(self, group_id=None):
        from euphorie.client.model import Group

        group_id = group_id or self.group_id
        return self.session.query(Group).filter(Group.group_id == group_id).first()

    def _do_csrf_check(self):
        # Call the check method of the transform.  This is called by
        # transformIterable, but that does too much.  We do need to prepare
        # the transform and the response a bit.
        resp = self.request.response
        resp.setHeader("Content-Type", "text/html")
        self.transform.site = self.portal

        # Do the check.
        self.transform.check()

    def _add_authenticator_to_request(self):
        self.request["_authenticator"] = createToken()


class IntegrationTests(BaseProtectTestCase):

    def test_start(self):
        # Test the start situation: there are no registered objects in the
        # current ZODB transaction.
        self.assertEqual(self.transform._registered_sql_objects(), [])
        self.assertEqual(self.transform._registered_objects(), [])

    def test_add_standard(self):
        item = self._createItem()
        self.session.add(item)
        self.assertEqual(self.transform._registered_sql_objects(), [item])
        # Note: we don't check _registered_objects here, as that depends
        # on whether CSRF protection for SQL is enabled or not.

    def test_add_with_flush(self):
        item = self._createItem()
        self.session.add(item)
        self.session.flush()
        self.assertEqual(self.transform._registered_sql_objects(), [item])


class FunctionalCSRFDisabledTests(BaseProtectTestCase, EuphorieFunctionalTestCase):
    """Functional test so we can do commits.

    Here we want to add an item, commit this, and edit or delete it.
    We explicitly DISABLE CSRF protection for SQLAlchemy writes in this test case.
    """

    def setUp(self):
        super().setUp()
        self.transform.euphorie_enable_csrf_protection_for_sql = False
        item = self._createItem()
        self.session.add(item)
        transaction.commit()
        self.group_id = item.group_id

    def test_start(self):
        self.assertEqual(self.transform._registered_sql_objects(), [])
        self.assertEqual(self.transform._registered_objects(), [])

    def test_edit_standard(self):
        # This is the start of a new transaction, so we need to get the item
        # fresh from the session.
        item = self._getItem()

        # Change a column.
        item.short_name = "changed"
        self.assertEqual(self.transform._registered_sql_objects(), [item])

        # Check that the change is really there.
        self.assertEqual(self._getItem().short_name, "changed")

        # Check that the change stays there after a full commit.
        transaction.commit()
        self.assertEqual(self._getItem().short_name, "changed")

    def test_edit_with_flush(self):
        item = self._getItem()
        item.short_name = "changed"
        self.session.flush()
        self.assertEqual(self.transform._registered_sql_objects(), [item])
        self.assertEqual(self._getItem().short_name, "changed")
        transaction.commit()
        self.assertEqual(self._getItem().short_name, "changed")

    def test_delete_standard(self):
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [item])

        # Check that the item is really deleted.
        self.assertIsNone(self._getItem())
        transaction.commit()
        self.assertIsNone(self._getItem())

    def test_delete_with_flush(self):
        item = self._getItem()
        self.session.delete(item)
        self.session.flush()
        self.assertEqual(self.transform._registered_sql_objects(), [item])
        self.assertIsNone(self._getItem())
        transaction.commit()
        self.assertIsNone(self._getItem())

    def test_delete_disable_csrf(self):
        """If we use the marker interface, no SQL objects are registered."""
        from euphorie.content.protect import IDisableCSRFProtectionForSQL

        alsoProvides(self.request, IDisableCSRFProtectionForSQL)
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [])

    def test_delete_check_get_without_authenticator(self):
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [item])

        # Actually do the csrf check.  A warning should be logged.
        with self.assertLogs(level="WARNING") as logged:
            self._do_csrf_check()
        self.assertTrue(logged.output)
        self.assertIn(
            "CSRF protection for SQLAlchemy changes is not enabled", logged.output[0]
        )

        # The write is accepted, because we have disabled SQL CSRF protection.
        resp = self.request.response
        self.assertEqual(resp.status, 200)

        # The transaction was not aborted, so the item no longer exists.
        self.assertIsNone(self._getItem())

    def test_delete_check_post_without_authenticator(self):
        self.request.REQUEST_METHOD = "POST"
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [item])

        # Actually do the csrf check.  A warning should be logged.
        with self.assertLogs(level="WARNING") as logged:
            self._do_csrf_check()
        self.assertTrue(logged.output)
        self.assertIn(
            "CSRF protection for SQLAlchemy changes is not enabled", logged.output[0]
        )

        # The write is accepted, because we have disabled SQL CSRF protection.
        resp = self.request.response
        self.assertEqual(resp.status, 200)

        # The transaction was not aborted, so the item no longer exists.
        self.assertIsNone(self._getItem())


class FunctionalCSRFEnabledTests(BaseProtectTestCase, EuphorieFunctionalTestCase):
    """Functional test so we can do commits.

    Here we want to add an item, commit this, and edit or delete it.
    We explicitly ENABLE CSRF protection for SQLAlchemy writes in this test case.
    """

    def setUp(self):
        super().setUp()
        self.transform.euphorie_enable_csrf_protection_for_sql = True
        item = self._createItem()
        self.session.add(item)
        transaction.commit()
        self.group_id = item.group_id

    def test_start(self):
        self.assertEqual(self.transform._registered_sql_objects(), [])
        self.assertEqual(self.transform._registered_objects(), [])

    def test_delete_check_get_without_authenticator(self):
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [item])

        # Actually do the csrf check.
        self._do_csrf_check()

        # The write on GET was detected, so we get redirected.
        resp = self.request.response
        self.assertEqual(resp.status, 302)
        self.assertIn("confirm-action", resp.getHeader("location"))

        # The transaction was aborted, so the item still exists.
        self.assertIsNotNone(self._getItem())

    def test_delete_check_get_with_authenticator(self):
        self._add_authenticator_to_request()
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [item])

        # Actually do the csrf check.
        self._do_csrf_check()

        # The write on GET was accepted, because we have an authenticator.
        resp = self.request.response
        self.assertEqual(resp.status, 200)

        # The transaction was not aborted, so the item no longer exists.
        self.assertIsNone(self._getItem())

    def test_delete_check_post_without_authenticator(self):
        self.request.REQUEST_METHOD = "POST"
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [item])

        # Actually do the csrf check.  We expect:
        # zExceptions.Forbidden: Form authenticator is invalid.
        with self.assertRaises(Forbidden):
            self._do_csrf_check()

        # The transaction was aborted, so the item still exists.
        self.assertIsNotNone(self._getItem())

    def test_delete_check_post_with_authenticator(self):
        self.request.REQUEST_METHOD = "POST"
        self._add_authenticator_to_request()
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [item])

        # Actually do the csrf check.
        self._do_csrf_check()

        # The write on POST was accepted, because we have an authenticator.
        resp = self.request.response
        self.assertEqual(resp.status, 200)

        # The transaction was not aborted, so the item no longer exists.
        self.assertIsNone(self._getItem())

    def test_delete_disable_csrf(self):
        from euphorie.content.protect import IDisableCSRFProtectionForSQL

        alsoProvides(self.request, IDisableCSRFProtectionForSQL)
        item = self._getItem()
        self.session.delete(item)
        self.assertEqual(self.transform._registered_sql_objects(), [])

        # Actually do the csrf check.
        self._do_csrf_check()

        # The write on GET is accepted.
        resp = self.request.response
        self.assertEqual(resp.status, 200)

        # The transaction was not aborted, so the item no longer exists.
        self.assertIsNone(self._getItem())

        # Rolling back the session or aborting the transaction brings back
        # the item.
        self.session.rollback()
        self.assertIsNotNone(self._getItem())



from ftw.upgrade import UpgradeStep


class Make_KeepMeLoggedIn_checkboxOptional(UpgradeStep):
    """Make 'Keep me logged in' checkbox optional."""

    def __call__(self):
        self.install_upgrade_profile()
